/**
 * @file          /kylin-license/src/daemon/dbus/dbus-license-object.cpp
 * @brief
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */

#include "src/daemon/dbus/dbus-license-object.h"
#include "lib/base/base.h"
#include "src/daemon/daemon.h"

namespace Daemon
{
#define KYLINSEC_OS_FILE "/etc/.kyinfo"

DBusLicenseObject::DBusLicenseObject(DBus::Connection &connection,
                                     std::shared_ptr<LicenseObject> license_object) : DBus::ObjectAdaptor(connection,
                                                                                                          fmt::format("{0}/{1}", LICENSE_OBJECT_OBJECT_PATH, license_object->get_name())),
                                                                                      license_object_(license_object)

{
    this->object_path_ = fmt::format("{0}/{1}", LICENSE_OBJECT_OBJECT_PATH, this->license_object_->get_name());
}

DBusLicenseObject::~DBusLicenseObject()
{
}

void DBusLicenseObject::init()
{
    this->license_object_->signal_license_info_changed().connect(sigc::mem_fun(this, &DBusLicenseObject::on_license_info_changed));
}

std::string DBusLicenseObject::GetLicense()
{
    KLOG_PROFILE("");

    auto license_info = this->license_object_->get_license_info();

    Json::Value ukey_value;
    Json::StreamWriterBuilder wbuilder;
    wbuilder["indentation"] = "";

    auto configuration = Daemon::get_instance()->get_configuration();
    ukey_value[LICENSE_JK_ACTIVATION_STATUS] = license_info.licensed;
    ukey_value[LICENSE_JK_SERVICE_STATUS] = license_info.service_status;
    ukey_value[LICENSE_JK_MACHINE_CODE] = license_info.machine_code;
    ukey_value[LICENSE_JK_EXPIRED_TIME] = Json::UInt64(license_info.expired_time);
    ukey_value[LICENSE_JK_FIRST_BOOT_TIME] = Json::UInt64(license_info.first_boot_time);
    ukey_value[LICENSE_JK_INSTALL_TYPE] = configuration->get_install_type();
    ukey_value[LICENSE_JK_IS_PG] = configuration->is_pg_version();
    ukey_value[LICENSE_JK_UPDATE_MODE] = license_info.update_mode;
    ukey_value[LICENSE_JK_SERVICE_MODE] = license_info.service_mode;
    ukey_value[LICENSE_JK_EXCEPTION_STATUS] = license_info.exception_status;
    ukey_value[LICENSE_JK_SERVICE_UID] = std::string();

    if (license_info.licensed == LicenseActivationStatus::LAS_ACTIVATED)
    {
        ukey_value[LICENSE_JK_ACTIVATION_CODE] = license_info.activation_code;
        ukey_value[LICENSE_JK_ACTIVATION_TYPE] = license_info.activation_type;
        ukey_value[LICENSE_JK_ACTIVATION_TIME] = license_info.activation_type == 0 ? Json::UInt64(0) : Json::UInt64(license_info.activation_time);
        ukey_value[LICENSE_JK_CUSTOMER_INFO] = license_info.customer_code;
        if (license_info.activation_type = LicenseActivationType::LICENSE_ACTIVATION_TYPE_ONLINE)
        {
            ukey_value[LICENSE_JK_SERVICE_UID] = license_info.service_uid;
        }
    }
    else
    {
        ukey_value[LICENSE_JK_ACTIVATION_CODE] = std::string();
        ukey_value[LICENSE_JK_ACTIVATION_TYPE] = LicenseActivationType::LICENSE_ACTIVATION_TYPE_UNKNOWN;
        ukey_value[LICENSE_JK_ACTIVATION_TIME] = 0;
        ukey_value[LICENSE_JK_CUSTOMER_INFO] = std::string();
    }

    return Json::writeString(wbuilder, ukey_value);
}

void DBusLicenseObject::ActivateByActivationCode(const std::string &activation_code)
{
    KLOG_PROFILE("Activation code: %s", activation_code.c_str());

    auto error_code = this->license_object_->activate_by_activation_code(activation_code);
    if (error_code != LicenseErrorCode::LEC_SUCCESS)
    {
        THROW_DBUSCXX_ERROR(this->dbus_error_message_, error_code);
    }
}

std::string DBusLicenseObject::ActivateByUsbKey()
{
    KLOG_PROFILE("");

    std::shared_ptr<UKeyDevice> ukey_device;

    auto error_code = this->license_object_->activate_by_usb_key(ukey_device);
    if (error_code != LicenseErrorCode::LEC_SUCCESS)
    {
        THROW_DBUSCXX_ERROR(this->dbus_error_message_, error_code);
    }
    else
    {
        Json::Value ukey_value;
        Json::StreamWriterBuilder wbuilder;
        wbuilder["indentation"] = "";

        const auto &high_address_area = ukey_device->get_high_address_area();

        ukey_value[LICENSE_JK_SUM_POINTS] = high_address_area.sum_points;
        ukey_value[LICENSE_JK_REMAINING_POINTS] = high_address_area.remaining_points;
        // 兼容之前版本
        ukey_value["almighty_authorized_points"] = high_address_area.sum_points;
        ukey_value["almighty_remaining_points"] = high_address_area.remaining_points;
        return Json::writeString(wbuilder, ukey_value);
    }
}

void DBusLicenseObject::ActivateOnLine(const std::string &host, const std::string &service_uid)
{
    KLOG_PROFILE("Host: %s", host.c_str());

    auto error_code = this->license_object_->activate_online(host, service_uid);
    if (error_code != LicenseErrorCode::LEC_SUCCESS)
    {
        THROW_DBUSCXX_ERROR(this->dbus_error_message_, error_code);
    }
}

void DBusLicenseObject::on_set_property(DBus::InterfaceAdaptor &interface,
                                        const std::string &property,
                                        const DBus::Variant &value)
{
}

std::string DBusLicenseObject::GetActivationCodeObject(const std::string &activation_code, const std::vector<std::string> &activation_objects)
{
    LicenseErrorCode error_code = LicenseErrorCode::LEC_SUCCESS;
    for (auto iter = activation_objects.begin(); iter != activation_objects.end(); iter++)
    {
        auto activation_code_fields = Base::LicenseUtils::parse_activation_code(activation_code, Base::LicenseUtils::generate_key(*iter), error_code);
        if (activation_code_fields)
        {
            return *iter;
        }
    }
    return Base::LicenseError::get_error_desc(error_code);
}

bool DBusLicenseObject::DeactiveLicense()
{
    return this->license_object_->deactive_license();
}

void DBusLicenseObject::on_license_info_changed()
{
    this->LicenseChanged(true);
}

}  // namespace Daemon
