#include "hardware-control.h"
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <algorithm>
#include <fstream>

namespace Daemon
{
#define CPU_CORE_CONFIG_FILE_PATH "/sys/devices/system/cpu/"
#define NET_DEVICE_PATH "/sys/class/net/"
#define CPU0_FILE_NAME "cpu0"

#define DISABLE_CPU_CORE "0\n"
#define ENABLE_CPU_CORE "1\n"

#define READ_BANDWIDTH_ERR -1

struct NetDeviceInfo
{
    NetDeviceInfo() : is_limit(false),
                      limit_way(0),
                      limit_speed(0) {}

    std::string net_device_name;
    bool is_limit;
    int limit_way;
    int limit_speed;
    std::vector<int> enabled_bandwidth;
};

struct CpuCoreInfo
{
    CpuCoreInfo() : is_enable(false) {}
    std::string cpu_core_name;
    bool is_enable;
};

enum NetDeviceLimitWay
{
    NET_DEVICE_UNLIMIT,
    NET_DEVICE_LIMIT_BY_ETHTOOL,
    NET_DEVICE_LIMIT_BY_TC,
    NET_DEVICE_LIMIT_END,
};

HardwareControl::HardwareControl(/* args */)
{
}

HardwareControl::~HardwareControl()
{
}

HardwareControl* HardwareControl::instance_ = NULL;
void HardwareControl::global_init()
{
    instance_ = new HardwareControl();
    instance_->init();
}

void HardwareControl::global_deinit()
{
    delete instance_;
}

void HardwareControl::init()
{
    this->read_cpu_core_info(this->cpu_cores_info_);
    this->read_net_devices_info(this->net_devices_info_);
}

void HardwareControl::limit_bandwidth(int bandwidth)
{
    for (auto iter = this->net_devices_info_.begin(); iter != this->net_devices_info_.end(); iter++)
    {
        // 防止多次重复限制导致的网卡故障
        if ((*iter)->is_limit == true && (*iter)->limit_speed == bandwidth)
        {
            continue;
        }

        auto limit_way = this->limit_net_device((*iter)->net_device_name, bandwidth);
        if (limit_way != NET_DEVICE_UNLIMIT)
        {
            (*iter)->limit_way = limit_way;
            (*iter)->is_limit = true;
            (*iter)->limit_speed = bandwidth;
        }
    }
}

void HardwareControl::limit_cpu_core(int core_number)
{
    // 需要预留两颗cpu给客户使用
    if (core_number > this->get_cpu_core_number() - 2)
    {
        KLOG_ERROR("Limit cpu core number must less than %d.", this->get_cpu_core_number() - 2);
        return;
    }

    auto disabled_core_number = this->get_disabled_cpu_number();

    if (disabled_core_number == core_number)
    {
        return;
    }
    //需要限制cpu核数量小于已限制cpu核数量
    else if (disabled_core_number > core_number)
    {
        auto enable_core_number = disabled_core_number - core_number;
        for (auto iter = this->cpu_cores_info_.begin(); iter != this->cpu_cores_info_.end(); iter++)
        {
            if (enable_core_number == 0)
            {
                return;
            }

            if (!(*iter)->is_enable && set_cpu_core_state((*iter)->cpu_core_name, true))
            {
                (*iter)->is_enable = true;
                enable_core_number--;
            }
        }
    }
    else
    {
        auto disable_core_number = core_number - disabled_core_number;
        for (auto iter = this->cpu_cores_info_.begin(); iter != this->cpu_cores_info_.end(); iter++)
        {
            if (disable_core_number == 0)
            {
                return;
            }
            if ((*iter)->is_enable && set_cpu_core_state((*iter)->cpu_core_name, false))
            {
                (*iter)->is_enable = false;
                disable_core_number--;
            }
        }
    }
}

void HardwareControl::unlimit_all_net_device()
{
    for (auto iter = this->net_devices_info_.begin(); iter != this->net_devices_info_.end(); iter++)
    {
        if ((*iter)->limit_way == NET_DEVICE_LIMIT_BY_ETHTOOL)
        {
            auto command_line = fmt::format("/usr/sbin/ethtool -s {0} speed {1} duplex full autoneg off",
                                            (*iter)->net_device_name,
                                            this->get_net_device_max_bandwidth((*iter)->net_device_name));
            std::string standard_error;
            std::string standard_output;

            Glib::spawn_command_line_sync(command_line, &standard_output, &standard_error);
            if (standard_error.length() > 0)
            {
                KLOG_WARNING("Unlimit device %s failed: %s", (*iter)->net_device_name.c_str(), standard_error.c_str());
            }
        }
        else if ((*iter)->limit_way == NET_DEVICE_LIMIT_BY_TC)
        {
            auto command_line = fmt::format("tc qdisc del dev {0} root tbf limit {1}Mbit burst {2}Mbit rate {3}Mbit",
                                            (*iter)->net_device_name,
                                            (*iter)->limit_speed,
                                            (*iter)->limit_speed,
                                            (*iter)->limit_speed);
            std::string standard_error;
            std::string standard_output;
            Glib::spawn_command_line_sync(command_line, &standard_output, &standard_error);
            if (standard_error.length() > 0)
            {
                KLOG_WARNING("Unlimit device %s failed!", (*iter)->net_device_name.c_str());
                continue;
            }
        }
        (*iter)->limit_way = NET_DEVICE_UNLIMIT;
        (*iter)->is_limit = false;
    }
}

void HardwareControl::enable_all_cpu_core()
{
    for (auto iter = this->cpu_cores_info_.begin(); iter != this->cpu_cores_info_.end(); iter++)
    {
        if (!(*iter)->is_enable)
        {
            if (this->set_cpu_core_state((*iter)->cpu_core_name, ENABLE_CPU_CORE))
            {
                (*iter)->is_enable = true;
            }
        }
    }
}

bool HardwareControl::is_cpu_core_limit()
{
    return this->get_disabled_cpu_number() > 0;
}

bool HardwareControl::is_bandwidth_limit()
{
    for (auto iter = this->net_devices_info_.begin(); iter != this->net_devices_info_.end(); iter++)
    {
        if ((*iter)->is_limit == true)
        {
            return true;
        }
    }
    return false;
}

int HardwareControl::limit_net_device(const std::string& net_dev_name, int bandwidth)
{
    if (this->is_support_bandwidth(net_dev_name, bandwidth))
    {
        auto command_line = fmt::format("/usr/sbin/ethtool -s {0} speed {1} duplex full autoneg off", net_dev_name, bandwidth);
        std::string standard_error;
        std::string standard_output;

        Glib::spawn_command_line_sync(command_line, &standard_output, &standard_error);
        if (standard_error.length() > 0)
        {
            KLOG_WARNING("%s", standard_error.c_str());
        }
    }

    // 判断是否设置成功,不成功使用“tc”命令重新限制
    if (this->get_net_device_bandwidth(net_dev_name) != bandwidth)
    {
        KLOG_INFO("use ethtool setting bandwidth failed, using tc try again!");

        auto command_line = fmt::format("/usr/sbin/tc qdisc add dev {0} root tbf limit {1}Mbit burst {2}Mbit rate {3}Mbit ",
                                        net_dev_name, bandwidth, bandwidth, bandwidth);
        std::string standard_error;
        std::string standard_output;
        Glib::spawn_command_line_sync(command_line, &standard_output, &standard_error);
        if (standard_error.length() > 0)
        {
            KLOG_WARNING("Use command \"tc\" set bandwidth failed: %s", standard_error.c_str());
            return NET_DEVICE_UNLIMIT;
        }
        else
        {
            return NET_DEVICE_LIMIT_BY_TC;
        }
    }
    else
    {
        return NET_DEVICE_LIMIT_BY_ETHTOOL;
    }
}

int HardwareControl::get_cpu_core_number()
{
    //此处加一是将cpu0加入，cpu0不受控制但需要计入核心总数
    return this->cpu_cores_info_.size() + 1;
}

int HardwareControl::get_net_device_bandwidth(const std::string& net_device_name)
{
    char bandwidth_buf[20];
    int bandwidth = READ_BANDWIDTH_ERR;

    auto fd = open((NET_DEVICE_PATH + net_device_name + "/speed").c_str(), O_RDONLY);
    if (-1 == fd)
    {
        KLOG_WARNING("Open file %s error: %s", (NET_DEVICE_PATH + net_device_name + "/speed").c_str(), strerror(errno));
    }
    else
    {
        int buf_length = read(fd, bandwidth_buf, sizeof(bandwidth_buf));
        if (-1 == buf_length)
        {
            KLOG_ERROR("Read file %s error: %s", (NET_DEVICE_PATH + net_device_name + "/speed").c_str(), strerror(errno));
        }
        else
        {
            bandwidth = atoi(bandwidth_buf);
        }
    }

    return bandwidth;
}

bool HardwareControl::set_cpu_core_state(const std::string& cpu_core_name, bool enable)
{
    auto cpu_control_file = CPU_CORE_CONFIG_FILE_PATH + cpu_core_name + "/online";
    if (Glib::file_test(cpu_control_file, Glib::FILE_TEST_EXISTS))
    {
        std::ofstream ofs(cpu_control_file, std::ios::in | std::ios::out);
        if (ofs)
        {
            enable ? ofs.write(ENABLE_CPU_CORE, 1) : ofs.write(DISABLE_CPU_CORE, 1);
            ofs.close();
            return true;
        }
        else
        {
            KLOG_WARNING("cpu %s can not be control: %s can't opend",
                         cpu_core_name.c_str(),
                         cpu_control_file.c_str());
            return false;
        }
    }
    return false;
}

bool HardwareControl::get_cpu_core_state(const std::string& cpu_core_name, bool& cpu_core_state)
{
    auto cpu_control_file = CPU_CORE_CONFIG_FILE_PATH + cpu_core_name + "/online";
    if (Glib::file_test(cpu_control_file, Glib::FILE_TEST_EXISTS))
    {
        std::ifstream ifs(cpu_control_file, std::ios::in | std::ios::out);
        if (ifs)
        {
            char buf[20] = {0};
            ifs.read(buf, sizeof(buf));
            auto state = std::string(buf);
            state == ENABLE_CPU_CORE ? cpu_core_state = true : cpu_core_state = false;
            ifs.close();
            return true;
        }
        else
        {
            KLOG_WARNING("cpu %s : %s can't opend.",
                         cpu_core_name.c_str(),
                         cpu_control_file.c_str());
            return false;
        }
    }
    else
    {
        KLOG_WARNING("cpu %s : %s file not exists.",
                     cpu_core_name.c_str(),
                     cpu_control_file.c_str());
        return false;
    }
}

int HardwareControl::get_disabled_cpu_number()
{
    int result = 0;
    for (auto iter = this->cpu_cores_info_.begin(); iter != this->cpu_cores_info_.end(); iter++)
    {
        if ((*iter)->is_enable == false)
        {
            result++;
        }
    }
    return result;
}

void HardwareControl::read_cpu_core_info(std::vector<std::shared_ptr<CpuCoreInfo>>& cpu_cores_info)
{
    try
    {
        Glib::Dir cpu_core_dir(CPU_CORE_CONFIG_FILE_PATH);

        while (true)
        {
            auto file_name = cpu_core_dir.read_name();
            if (file_name.empty())
            {
                break;
            }

            // cpu0具有特殊性，部分机器无法控制，cpu0状态永久enable，并计入enable核心中
            if (file_name == CPU0_FILE_NAME)
            {
                continue;
            }

            if (Glib::file_test(CPU_CORE_CONFIG_FILE_PATH + file_name, Glib::FileTest::FILE_TEST_IS_DIR))
            {
                if ("cpu" == file_name.substr(0, 3) && (file_name[3] >= '0' && file_name[3] <= '9'))
                {
                    auto cpu_core_info = std::make_shared<CpuCoreInfo>();
                    cpu_core_info->cpu_core_name = file_name;
                    if (!this->get_cpu_core_state(file_name, cpu_core_info->is_enable))
                    {
                        continue;
                    }

                    cpu_cores_info.push_back(cpu_core_info);
                }
            }
        }
    }
    catch (const Glib::Error& e)
    {
        KLOG_WARNING("%s", e.what().c_str());
    }
}

void HardwareControl::read_net_devices_info(std::vector<std::shared_ptr<NetDeviceInfo>>& net_devices_info)
{
    try
    {
        Glib::Dir net_devices_dir(NET_DEVICE_PATH);

        while (true)
        {
            auto file_name = net_devices_dir.read_name();
            if (file_name.empty())
            {
                break;
            }
            if (Glib::file_test(NET_DEVICE_PATH + file_name, Glib::FileTest::FILE_TEST_IS_DIR) &&
                Glib::file_test(NET_DEVICE_PATH + file_name + "/device", Glib::FileTest::FILE_TEST_EXISTS))
            {
                auto device_info = std::make_shared<NetDeviceInfo>();
                device_info->net_device_name = file_name;
                device_info->is_limit = false;
                device_info->limit_way = NET_DEVICE_UNLIMIT;
                device_info->enabled_bandwidth = this->get_enabled_bandwidth(file_name);
                // wifi设备不做限制
                if (!this->is_wifi_device(file_name))
                {
                    net_devices_info.push_back(device_info);
                }
            }
        }
    }
    catch (const Glib::Error& e)
    {
        KLOG_WARNING("%s", e.what().c_str());
    }
}

bool HardwareControl::is_wifi_device(const std::string& device_name)
{
    return (Glib::file_test(NET_DEVICE_PATH + device_name + "/wireless", Glib::FILE_TEST_EXISTS) ||
            Glib::file_test(NET_DEVICE_PATH + device_name + "/phy80211", Glib::FILE_TEST_EXISTS));
}

bool HardwareControl::reset_net_device(const std::string& device_name)
{
    auto command_line = fmt::format("ifconfig {0} down", device_name);
    std::string standard_error;
    std::string standard_output;
    Glib::spawn_command_line_sync(command_line, &standard_output, &standard_error);
    if (standard_error.length() > 0)
    {
        KLOG_ERROR("Command execution failed: %s", command_line.c_str());
        return false;
    }

    command_line = fmt::format("ifconfig {0} up", device_name);
    if (standard_error.length() > 0)
    {
        KLOG_ERROR("Command execution failed: %s", command_line.c_str());
        return false;
    }

    return true;
}

std::vector<int> HardwareControl::get_enabled_bandwidth(const std::string& device_name)
{
    std::vector<int> result;

    std::string standard_error;
    std::string standard_output;
    auto command_line = fmt::format("/usr/sbin/ethtool {0}", device_name);
    Glib::spawn_command_line_sync(command_line, &standard_output, &standard_error);
    if (standard_error.length() > 0)
    {
        KLOG_WARNING("Use ethtool get device: %s failed.", device_name.c_str());
        return result;
    }

    // 去除'/n' '/r'
    auto preliminary_str_list = Base::StrUtils::split_lines(standard_output);

    std::string no_lines_str;
    // 将字符串重新拼接
    for (auto iter = preliminary_str_list.begin(); iter != preliminary_str_list.end(); iter++)
    {
        no_lines_str += *iter;
        no_lines_str += " ";
    }

    // 通过' '分割字符串
    auto splite_str_list = Base::StrUtils::split_with_char(no_lines_str, ' ');

    for (auto iter = splite_str_list.begin(); iter != splite_str_list.end(); iter++)
    {
        if (iter->find("baseT/Full") == iter->npos)
        {
            continue;
        }
        auto number_list = Base::StrUtils::split_with_string(*iter, "baseT/Full");

        for (auto iter = number_list.begin(); iter != number_list.end(); iter++)
        {
            if (Base::StrUtils::is_number(*iter))
            {
                if (std::atol(iter->c_str()) == 0)
                {
                    KLOG_DEBUG("number_list: %s", (*iter).c_str());
                }
                result.push_back(std::atol(iter->c_str()));
            }
        }
    }
    return result;
}

int HardwareControl::get_net_device_max_bandwidth(const std::string& device_name)
{
    int result = 0;

    for (auto iter = this->net_devices_info_.begin(); iter != this->net_devices_info_.end(); iter++)
    {
        if ((*iter)->net_device_name == device_name)
        {
            result = *std::max_element((*iter)->enabled_bandwidth.begin(), (*iter)->enabled_bandwidth.end());
        }
    }
    return result;
}

bool HardwareControl::is_support_bandwidth(const std::string& device_name, int bandwidth)
{
    for (auto iter = this->net_devices_info_.begin(); iter != this->net_devices_info_.end(); iter++)
    {
        if ((*iter)->net_device_name == device_name)
        {
            for (int i = 0; i < (*iter)->enabled_bandwidth.size(); i++)
            {
                if ((*iter)->enabled_bandwidth[i] == bandwidth)
                {
                    return true;
                }
            }
        }
    }
    return false;
}

}  // namespace Daemon