/**
 * @file          /ks-license/src/daemon/license-object-hw-joint.cpp
 * @brief
 * @author        xiaowen <xiaowen@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */

#include "license-object-os-hw-joint.h"

namespace Daemon
{
#define REPO_FILE_NAME "/etc/yum.repos.d/KylinSec.repo"
#define REPO_GROUP_UPDATE "update"
LicenseObjectOSHWJoint::LicenseObjectOSHWJoint(const std::string &name) : LicenseObjectOS(name)
{
    this->yum_repo_control_ = std::make_shared<YumRepoControl>(REPO_FILE_NAME);
    this->yum_repo_control_->insert_repo_group(REPO_GROUP_UPDATE);
    this->yum_repo_control_->signal_certificate_file_deleted().connect(sigc::mem_fun<void>(this, &LicenseObjectOSHWJoint::update_certificate_file));
    this->update_certificate_file();

    auto timeout = Glib::MainContext::get_default()->signal_timeout();
    this->certificate_timer_ = timeout.connect_seconds(sigc::mem_fun(this, &LicenseObjectOSHWJoint::timing_check_certificate), LICENSE_CHECK_INTERVAL);

    // 激活解除限制
    this->signal_activation_success().connect(sigc::mem_fun(this, &LicenseObjectOSHWJoint::on_activation_success));
}

LicenseObjectOSHWJoint::~LicenseObjectOSHWJoint()
{
}

void LicenseObjectOSHWJoint::update_certificate_file()
{
#ifdef ENABLE_SERVER
    auto license_info = this->get_license_info();

    // 在订阅模式下质保期已过则不允许yum update操作，所以订阅模式过期则不再更新证书
    if (license_info.licensed == LicenseActivationStatus::LAS_ACTIVATED &&
        license_info.update_mode == LicenseUpdateMode::LICENSE_UPDATE_MODE_SUBSCRIBE &&
        license_info.service_status != LicenseServiceStatus::LICENSE_SERVICE_STATUS_UNEXPIRED)
    {
        return;
    }

    if (license_info.licensed == LicenseActivationStatus::LAS_ACTIVATED)
    {
        if (this->yum_repo_control_->download_certificate_file())
        {
            auto license_object_group = this->configuration_->get_license_object_group(this->get_name());
            auto certifacate_expired_timestamp = time(NULL) + DEFAULT_VALIDITY_PERIOD * DAY_IN_SECONDS;
            license_object_group->certificate_expired_time = Base::TimeUtils::timestamp2str(certifacate_expired_timestamp);
            this->configuration_->set_license_object_group(license_object_group);
        }
    }
#endif
}

bool LicenseObjectOSHWJoint::is_updating_certificate()
{
#ifdef ENABLE_SERVER
    auto license_object_group = this->configuration_->get_license_object_group(this->get_name());
    RETURN_VAL_IF_TRUE(!license_object_group || license_object_group->certificate_expired_time.empty(), true);
    auto certificate_expired_timestamp = Base::TimeUtils::str2timestamp(license_object_group->certificate_expired_time.c_str());
    // 由于证书只能精确到天，所以提前一天更新证书
    RETURN_VAL_IF_TRUE(certificate_expired_timestamp < 3600 * 24 + time(NULL), true);
#endif
    return false;
}

bool LicenseObjectOSHWJoint::timing_check_certificate()
{
    // 证书快要过期的情况下更新证书
    if (this->is_updating_certificate())
    {
        this->update_certificate_file();
    }

    return true;
}

void LicenseObjectOSHWJoint::on_activation_success()
{
    KLOG_PROFILE("");
    this->update_certificate_file();
}

}  // namespace Daemon