/**
 * @file          /ks-license/src/daemon/license-object-os-pg.cpp
 * @brief
 * @author        xiaowen <xiaowen@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */

#include "license-object-os-pg.h"
#include <fcntl.h>
#include <sys/ioctl.h>
#include <fstream>

namespace Daemon
{
#define EXPIRY_DAYS 0
#define SERIOUS_EXPIRY_DAYS MONTH_IN_DAYS
#define REPO_FILE_NAME "/etc/yum.repos.d/KylinSec.repo"
#define REPO_GROUP_EVERYTHING "everything"
#define REPO_GROUP_EPOL "EPOL"
#define REPO_GROUP_EXTRA "extra"
#define REPO_GROUP_CDROM "CDROM"
#define REPO_GROUP_UPDATE "update"
#define REPO_GROUP_RELEASE_UPDATE "release-update"
#define UNLIMIT_FILE "unlimit"

LicenseObjectOSPG::LicenseObjectOSPG(const std::string& name) : LicenseObjectOS(name)
{
    this->hardware_control = HardwareControl::get_instance();

    this->expiry_signal_ = std::make_shared<sigc::signal<void>>();
    this->serious_expiry_signal_ = std::make_shared<sigc::signal<void>>();

    auto license_info = this->get_license_info();
    if ((license_info.licensed == LicenseActivationStatus::LAS_UNACTIVATED ||
         license_info.update_mode == LicenseUpdateMode::LICENSE_UPDATE_MODE_SUBSCRIBE) &&
        !this->check_unlimit_file_validity())
    {
        this->regist_license_state(EXPIRY_DAYS, this->expiry_signal_);
        this->regist_license_state(SERIOUS_EXPIRY_DAYS, this->serious_expiry_signal_);
    }

    this->yum_repo_control_ = std::make_shared<YumRepoControl>(REPO_FILE_NAME);
    this->yum_repo_control_->insert_repo_group(REPO_GROUP_EVERYTHING);
    this->yum_repo_control_->insert_repo_group(REPO_GROUP_EPOL);
    this->yum_repo_control_->insert_repo_group(REPO_GROUP_EXTRA);
    this->yum_repo_control_->insert_repo_group(REPO_GROUP_UPDATE);

    // 开启后门防止特殊情况导致的激活状态变化限制硬件
    this->unlimit_monitor_ = Base::FileUtils::make_monitor_file(std::string(KL_INSTALL_DATADIR) + "/" + UNLIMIT_FILE,
                                                                sigc::mem_fun(this, &LicenseObjectOSPG::on_unlimit_file_changed));

    this->yum_repo_control_->signal_certificate_file_deleted().connect(sigc::mem_fun<void>(this, &LicenseObjectOSPG::update_certificate_file));
    this->expiry_signal_->connect(sigc::mem_fun(this, &LicenseObjectOSPG::expiry_limit));
    this->serious_expiry_signal_->connect(sigc::mem_fun(this, &LicenseObjectOSPG::serious_expiry_limit));
    this->signal_activation_success().connect(sigc::mem_fun(this, &LicenseObjectOSPG::on_activation_success));
    this->signal_check_license_state().connect(sigc::mem_fun(this, &LicenseObjectOSPG::on_check_license_state));

    auto timeout = Glib::MainContext::get_default()->signal_timeout();
    this->certificate_timer_ = timeout.connect_seconds(sigc::mem_fun(this, &LicenseObjectOSPG::timing_check_certificate), LICENSE_CHECK_INTERVAL);

    // 程序启检测一次
    this->on_check_license_state();
    // 程序启动拉取一次证书文件
    this->update_certificate_file();
}

LicenseObjectOSPG::~LicenseObjectOSPG()
{
}

void LicenseObjectOSPG::expiry_limit()
{
    KLOG_PROFILE("");
    KLOG_INFO("Limit cpu core as half, limit bandwidth 10Mb/s");
    this->hardware_control->limit_cpu_core(this->hardware_control->get_cpu_core_number() / 2);
    this->hardware_control->limit_bandwidth(10);
}

void LicenseObjectOSPG::serious_expiry_limit()
{
    KLOG_PROFILE("");
    KLOG_INFO("Enable only 2 cpu cores, limit bandwidth 10Mb/s");
    this->hardware_control->limit_cpu_core(this->hardware_control->get_cpu_core_number() - 2);
    this->hardware_control->limit_bandwidth(10);
}

void LicenseObjectOSPG::on_activation_success()
{
    KLOG_PROFILE("");

    // 订阅模式下需要进行过期限制
    auto license_info = this->get_license_info();
    this->clear_license_state();
    if (license_info.update_mode == LicenseUpdateMode::LICENSE_UPDATE_MODE_SUBSCRIBE)
    {
        this->regist_license_state(EXPIRY_DAYS, this->expiry_signal_);
        this->regist_license_state(SERIOUS_EXPIRY_DAYS, this->serious_expiry_signal_);
    }

    this->unlimit_hardware();
    this->update_certificate_file();
}

void LicenseObjectOSPG::unlimit_hardware()
{
    if (this->hardware_control->is_bandwidth_limit())
    {
        this->hardware_control->unlimit_all_net_device();
        KLOG_INFO("unlimit all net devices.");
    }

    if (this->hardware_control->is_cpu_core_limit())
    {
        this->hardware_control->enable_all_cpu_core();
        KLOG_INFO("enable all cpu core.");
    }
}

void LicenseObjectOSPG::on_check_license_state()
{
    this->unlimit_hardware();
}

void LicenseObjectOSPG::update_certificate_file()
{
#ifdef ENABLE_SERVER
    auto license_info = this->get_license_info();

    // 在订阅模式下质保期已过则不允许yum update操作，所以订阅模式过期则不再更新证书
    if (license_info.licensed == LicenseActivationStatus::LAS_ACTIVATED &&
        license_info.update_mode == LicenseUpdateMode::LICENSE_UPDATE_MODE_SUBSCRIBE &&
        license_info.service_status != LicenseServiceStatus::LICENSE_SERVICE_STATUS_UNEXPIRED)
    {
        return;
    }
    // 试用期内或者已激活情况下拉取证书
    if (license_info.expired_time < time(NULL) || license_info.licensed == LicenseActivationStatus::LAS_ACTIVATED)
    {
        if (this->yum_repo_control_->download_certificate_file())
        {
            auto license_object_group = this->configuration_->get_license_object_group(this->get_name());
            auto certifacate_expired_timestamp = time(NULL) + DEFAULT_VALIDITY_PERIOD * DAY_IN_SECONDS;
            license_object_group->certificate_expired_time = Base::TimeUtils::timestamp2str(certifacate_expired_timestamp);
            this->configuration_->set_license_object_group(license_object_group);
        }
    }
#endif
}

bool LicenseObjectOSPG::is_updating_certificate()
{
#ifdef ENABLE_SERVER
    auto license_object_group = this->configuration_->get_license_object_group(this->get_name());
    RETURN_VAL_IF_TRUE(!license_object_group || license_object_group->certificate_expired_time.empty(), true);
    auto certificate_expired_timestamp = Base::TimeUtils::str2timestamp(license_object_group->certificate_expired_time.c_str());
    // 由于证书只能精确到天，所以提前一天更新证书
    RETURN_VAL_IF_TRUE(certificate_expired_timestamp < 3600 * 24 + time(NULL), true);
#endif
    return false;
}

bool LicenseObjectOSPG::timing_check_certificate()
{
    // 证书快要过期的情况下更新证书
    if (this->is_updating_certificate())
    {
        this->update_certificate_file();
    }

    return true;
}

void LicenseObjectOSPG::on_unlimit_file_changed(const Glib::RefPtr<Gio::File>& file,
                                                const Glib::RefPtr<Gio::File>& other_file,
                                                Gio::FileMonitorEvent event_type)
{
    KLOG_PROFILE("");
    if (event_type == Gio::FILE_MONITOR_EVENT_CREATED)
    {
        if (check_unlimit_file_validity())
        {
            this->clear_license_state();
            this->unlimit_hardware();
        }
    }
}

bool LicenseObjectOSPG::check_unlimit_file_validity()
{
    RETURN_VAL_IF_FALSE(Glib::file_test(std::string(KL_INSTALL_DATADIR) + "/" + UNLIMIT_FILE, Glib::FILE_TEST_EXISTS), false)

    auto machine_code = Base::StrUtils::trim(Glib::file_get_contents(std::string(KL_INSTALL_DATADIR) + "/" + UNLIMIT_FILE));
    if (this->get_license_info().machine_code == machine_code)
    {
        auto license_object_group = this->configuration_->get_license_object_group(this->name_);
        return license_object_group->once_active == true;
    }

    return false;
}

}  // namespace Daemon