/**
 * @file          /ks-license/src/daemon/license-object.cpp
 * @brief
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */

#include "src/daemon/license-object.h"

#ifdef ENABLE_SERVER
#include <grpc/grpc.h>
#include <grpcpp/client_context.h>
#include <grpcpp/create_channel.h>
#include <license.grpc.pb.h>
#include <license.pb.h>
#endif

#include <fmt/format.h>
#include <glib/gi18n.h>
#include <sys/stat.h>
#include <sys/sysinfo.h>
#include "lib/base/base.h"
#include "src/daemon/daemon.h"
#include "src/daemon/license-configuration.h"
#include "src/daemon/license-machine.h"
#include "src/daemon/ukey-manager.h"

namespace Daemon
{
#define DESKTOP_TYPE_MEMORY_SIZE_LIMIT (40UL * 1024UL * 1024UL * 1024UL)
/*客户端和服务端时间最大差异（秒)
 *该时间必须要大于LICENSE_CHECK_INTERVAL
 *否则会同时出现服务器客户端时间差过大和无法与服务器连接两种异常状态
 */
#define CS_TIMESTAMP_MAX_DIFF (60 * 3)
#define KYLINSEC_OS_FILE "/etc/.kyinfo"
#define MAX_EXPIRY_TIME "2100-01-02"

LicenseInfo::LicenseInfo() : licensed(LicenseActivationStatus::LAS_UNACTIVATED),
                             service_status(LicenseServiceStatus::LICENSE_SERVICE_STATUS_INVALID),
                             expired_time(0),
                             activation_type(LicenseActivationType::LICENSE_ACTIVATION_TYPE_UNKNOWN),
                             first_boot_time(time(NULL)),
                             activation_time(0),
                             update_mode(LicenseUpdateMode::LICENSE_UPDATE_MODE_NONE),
                             service_mode(LicenseServiceMode::LICENSE_SERVICE_MODE_NONE),
                             exception_status(LicenseExceptionStatus::LICENSE_EXCEPTION_STATUS_NONE)
{
}

LicenseObject::LicenseObject(const std::string &name) : name_(name)
{
    this->configuration_ = Daemon::get_instance()->get_configuration();
    this->key_ = Base::LicenseUtils::generate_key(name);
    // 首次开启忽略客户端服务器时间异常
    this->server_timestamp_ = 0;

    // 每10分钟检查一次授权，timing_check_license为虚函数，所以使用lambda进行封装
    auto timeout = Glib::MainContext::get_default()->signal_timeout();
    this->timing_timer_ = timeout.connect_seconds(sigc::mem_fun(this, &LicenseObject::timing_check_license),
                                                  LICENSE_CHECK_INTERVAL);

    // 程序启动时检测一次
    this->timing_check_license();

    this->configuration_->signal_group_changed().connect(sigc::mem_fun(this, &LicenseObject::on_license_group_changed));
    this->signal_license_info_changed().connect(sigc::mem_fun(this, &LicenseObject::on_license_info_changed));
    this->signal_activation_success().connect(sigc::mem_fun(this, &LicenseObject::on_activation_success));
}

LicenseObject::~LicenseObject()
{
}

bool LicenseObject::timeout_check_license()
{
    this->timing_check_license();
    return false;
}

bool LicenseObject::timing_check_license()
{
    KLOG_PROFILE("");

    this->update_license_info();

    // 如果即将过期，则添加一个定时器进行检测和更新授权状态
    auto license_info = this->get_license_info();
    auto expired_remaining_time = license_info.expired_time - time(NULL);
    KLOG_DEBUG("Expired remaining: %d.", expired_remaining_time);
    if (expired_remaining_time > 0 && expired_remaining_time < LICENSE_CHECK_INTERVAL &&
        !this->expired_timer_)
    {
        KLOG_DEBUG("The license will expired after %d seconds.", expired_remaining_time);
        auto timeout = Glib::MainContext::get_default()->signal_timeout();
        this->expired_timer_ = timeout.connect_seconds(sigc::mem_fun(this, &LicenseObject::timeout_check_license),
                                                       expired_remaining_time + 3);
    }

    // 如果是在线激活，需要跟激活服务器保持心跳连接
    if (!license_info.auth_server.empty())
    {
        if (license_info.licensed == LicenseActivationStatus::LAS_UNACTIVATED)
        {
            KLOG_INFO("Reactive !");
            this->activate_online_with_config();
        }

        if (license_info.licensed == LicenseActivationStatus::LAS_ACTIVATED &&
            !this->heartbeat_timer_.connected())
        {
            auto timeout = Glib::MainContext::get_default()->signal_timeout();
            this->heartbeat_timer_ = timeout.connect_seconds(sigc::mem_fun(this, &LicenseObject::on_heartbeat_connection), HEART_BEAT_INTERVAL);
        }
    }

    return true;
}

LicenseInfo LicenseObject::get_license_info()
{
    KLOG_PROFILE("");

    LicenseErrorCode error_code;
    LicenseInfo result;

    auto license_common_group = this->configuration_->get_license_common_group();
    auto license_object_group = this->configuration_->get_license_object_group(this->name_);

    std::shared_ptr<Base::ActivationCodeFields> activation_code_fileds;
    do
    {
        BREAK_IF_FALSE(license_object_group);
        result.auth_server = license_object_group->auth_server;

        BREAK_IF_FALSE(license_common_group);
        result.first_boot_time = Base::TimeUtils::str2timestamp(license_common_group->first_boot_time.c_str());

        BREAK_IF_TRUE(license_object_group->activation_code.empty());

        activation_code_fileds = Base::LicenseUtils::parse_activation_code(license_object_group->activation_code, this->key_, error_code);
        if (!activation_code_fileds)
        {
            auto error = LICENSE_ERROR2STR(error_code);
            KLOG_WARNING("%s", error.c_str());
            break;
        }

        auto machine_code = this->get_machine_code();
        if (!this->compare_machine_code(machine_code, activation_code_fileds->machine_code))
        {
            KLOG_WARNING("Failed to match machine code.");
            break;
        }

        /* 这里使用计算出来的机器码，而不是通过授权码解密出来的机器码，因为解密出来的机器码可能是通过旧的计算方式计算出来的。
             例如以前的服务器版本机器码是02开头，新的计算方式是0D开头，但这两个机器码都是可以在新版本中使用的*/
        result.machine_code = machine_code;
        result.service_uid = license_object_group->service_uid;

        // 记录在线激活异常状态
        if (license_object_group->activation_type == LicenseActivationType::LICENSE_ACTIVATION_TYPE_ONLINE &&
            license_object_group->keep_alive_deadline > 0)
        {
            // 判断异常码不等于0，且是虚拟机，认为未激活状态
            result.exception_status = this->get_exception_status(license_object_group->keep_alive_deadline);
            if (Base::LicenseUtils::is_virtual_machine() &&
                result.exception_status != LicenseExceptionStatus::LICENSE_EXCEPTION_STATUS_NONE)
            {
                break;
            }
        }

        result.licensed = LicenseActivationStatus::LAS_ACTIVATED;
    } while (false);

    this->fill_license_info(result, activation_code_fileds);

    return result;
}

std::string LicenseObject::get_machine_code()
{
    return Daemon::get_instance()->get_machine()->get_machine_code();
}

LicenseErrorCode LicenseObject::check_activation_code(const std::string &activation_code, LicenseActivationType activation_type)
{
    KLOG_PROFILE("Activation code: %s.", activation_code.c_str());

    std::string license_machine_code;
    LicenseErrorCode error_code = LicenseErrorCode::LEC_SUCCESS;

    auto machine_code = this->get_machine_code();
    auto activation_code_fileds = Base::LicenseUtils::parse_activation_code(activation_code, this->key_, error_code);
    RETURN_VAL_IF_FALSE(activation_code_fileds, error_code);

    if (activation_code_fileds->activation_type != LicenseActivationType::LICENSE_ACTIVATION_TYPE_UNKNOWN &&
        activation_code_fileds->activation_type != activation_type)
    {
        return LicenseErrorCode::LEC_LICENSE_ACTIVATION_CODE_TYPE_ERROR;
    }

    if (!this->compare_machine_code(machine_code, activation_code_fileds->machine_code))
    {
        return LicenseErrorCode::LEC_UNMATCH_MACHINE_CODE;
    }

    RETURN_VAL_IF_ERROR(Base::LicenseUtils::unformat_machine_code(activation_code_fileds->machine_code, license_machine_code));
    RETURN_VAL_IF_FALSE(Base::StrUtils::is_nature_str(activation_code_fileds->customer_info), LicenseErrorCode::LEC_CUSTOMER_INFO_INVALID);

    if (activation_code_fileds->expiry_timestamp > Base::TimeUtils::str2timestamp(MAX_EXPIRY_TIME))
    {
        return LicenseErrorCode::LEC_EXPIRED_DATE_LARGE;
    }

    struct sysinfo info;
    // 用户可能在服务器机器上安装桌面版本的系统，这里根据内存大小进行校验，服务器机器的内存一般大于40G
    // 限制桌面版本只能安装在40G内存以下的机器上
    if (!sysinfo(&info))
    {
        if (license_machine_code[0] == LICENSE_INSTALL_TYPE_DESKTOP &&
            info.totalram >= DESKTOP_TYPE_MEMORY_SIZE_LIMIT)
        {
            return LicenseErrorCode::LEC_INSTALL_TYPE_UNMATCH_MACHINE;
        }
    }

    if (activation_code_fileds->expiry_timestamp < time(NULL))
    {
        return LicenseErrorCode::LEC_LICENSE_EXPIRED;
    }
    return LicenseErrorCode::LEC_SUCCESS;
}

std::shared_ptr<Base::ActivationCodeFields> LicenseObject::check_activation_code(const std::string &activation_code,
                                                                                 LicenseErrorCode &error_code)
{
    KLOG_PROFILE("Activation code: %s.", activation_code.c_str());

    std::string license_machine_code;
    error_code = LicenseErrorCode::LEC_SUCCESS;

    auto machine_code = this->get_machine_code();
    auto activation_code_fields = Base::LicenseUtils::parse_activation_code(activation_code, this->key_, error_code);
    RETURN_VAL_IF_FALSE(activation_code_fields, std::shared_ptr<Base::ActivationCodeFields>());
    do
    {
        if (activation_code_fields->activation_type < LicenseActivationType::LICENSE_ACTIVATION_TYPE_UNKNOWN &&
            activation_code_fields->activation_type > LicenseActivationType::LICENSE_ACTIVATION_TYPE_ONLINE)
        {
            error_code = LicenseErrorCode::LEC_LICENSE_ACTIVATION_CODE_TYPE_ERROR;
            break;
        }

        if (!this->compare_machine_code(machine_code, activation_code_fields->machine_code))
        {
            error_code = LicenseErrorCode::LEC_UNMATCH_MACHINE_CODE;
            break;
        }

        error_code = Base::LicenseUtils::unformat_machine_code(activation_code_fields->machine_code, license_machine_code);
        if (LicenseErrorCode::LEC_SUCCESS != error_code)
        {
            break;
        }

        struct sysinfo info;
        // 用户可能在服务器机器上安装桌面版本的系统，这里根据内存大小进行校验，服务器机器的内存一般大于40G
        // 限制桌面版本只能安装在40G内存以下的机器上
        if (!sysinfo(&info))
        {
            if (license_machine_code[0] == LICENSE_INSTALL_TYPE_DESKTOP &&
                info.totalram >= DESKTOP_TYPE_MEMORY_SIZE_LIMIT)
            {
                error_code = LicenseErrorCode::LEC_INSTALL_TYPE_UNMATCH_MACHINE;
                break;
            }
        }

        if (activation_code_fields->expiry_timestamp < time(NULL))
        {
            error_code = LicenseErrorCode::LEC_LICENSE_EXPIRED;
            break;
        }
    } while (false);

    if (error_code == LicenseErrorCode::LEC_SUCCESS)
    {
        return activation_code_fields;
    }
    else
    {
        return std::shared_ptr<Base::ActivationCodeFields>();
    }
}

LicenseErrorCode LicenseObject::activate_by_activation_code(const std::string &activation_code)
{
    return this->activate_with_type(activation_code, LicenseActivationType::LICENSE_ACTIVATION_TYPE_MANNUAL);
}

LicenseErrorCode LicenseObject::activate_online(const std::string &host, const std::string &service_uid)
{
    KLOG_PROFILE("host: %s", host.c_str());

#ifdef ENABLE_SERVER
    grpc::ClientContext context;
    ::ActivationReq req;
    ::ActivationRsp rsp;
    auto channel = grpc::CreateChannel(fmt::format("{0}:{1}", host, DEFAULT_LISTEN_PORT), grpc::InsecureChannelCredentials());
    auto stub = License::NewStub(channel);

    auto current_timestamp = time(NULL);

    gpr_timespec deadline;
    deadline.tv_sec = LICENSE_CLIENT_TIMEOUT_SEC;
    deadline.tv_nsec = 0;
    deadline.clock_type = GPR_TIMESPAN;

    context.set_deadline(deadline);
    req.set_activation_object_name(this->name_);
    req.set_machine_code(this->get_machine_code());
    req.set_service_uid(service_uid);
    req.set_machine_type(Base::LicenseUtils::is_virtual_machine() ? LicenseMachineType::LICENSE_MACHINE_TYPE_VIRTUAL : LicenseMachineType::LICENSE_MACHINE_TYPE_PHYSICAL);
    req.set_customization_customer(RESTRICTION_POLICY);
    req.set_timestamp(uint32_t(current_timestamp));
    auto status = stub->GetActivationCode(&context, req, &rsp);

    if (status.ok())
    {
        this->server_timestamp_ = time_t(rsp.timestamp());

        if (rsp.error_code() == LicenseErrorCode::LEC_SUCCESS)
        {
            auto retval = this->activate_with_type(rsp.activation_code(), LicenseActivationType::LICENSE_ACTIVATION_TYPE_ONLINE);
            RETURN_VAL_IF_FALSE(retval == LicenseErrorCode::LEC_SUCCESS, retval);
        }
        else
        {
            return LicenseErrorCode(rsp.error_code());
        }
    }
    else
    {
        this->server_timestamp_ = 0;
        KLOG_WARNING("Failed to call GetActivationCode, status code: %d.", int32_t(status.error_code()));
        return LicenseErrorCode::LEC_SERVER_RESPONSE_ERROR;
    }

    auto license_object_group = this->configuration_->get_license_object_group(this->name_);
    RETURN_VAL_IF_FALSE(license_object_group, LicenseErrorCode::LEC_COMMON_NULL_POINTER);
    license_object_group->auth_server = host;
    license_object_group->keep_alive_deadline = (rsp.keep_alive_time() == 0) ? 0 : rsp.timestamp() + rsp.keep_alive_time();
    license_object_group->service_uid = service_uid;
    RETURN_VAL_IF_FALSE(this->configuration_->set_license_object_group(license_object_group), LEC_SYNC_TO_FILE_FAILED);

    return LicenseErrorCode::LEC_SUCCESS;
#else
    return LicenseErrorCode::LEC_LICENSE_NOT_SUPPORT_ONLINE_ACTIVATION;
#endif
}

LicenseErrorCode LicenseObject::activate_by_usb_key(std::shared_ptr<UKeyDevice> &used_ukey_device)
{
    KLOG_PROFILE("");

    std::string activation_code;
    auto machine_code = this->get_machine_code();

    used_ukey_device = UKeyManager::get_instance()->get_ukey(this->name_);
    RETURN_VAL_IF_FALSE(used_ukey_device, LicenseErrorCode::LEC_UKEY_NOT_FOUND);

    RETURN_VAL_IF_ERROR(used_ukey_device->generate_activation_code(machine_code, activation_code));
    RETURN_VAL_IF_ERROR(this->activate_with_type(activation_code, LicenseActivationType::LICENSE_ACTIVATION_TYPE_OFFLINE));
    return LicenseErrorCode::LEC_SUCCESS;
}

LicenseErrorCode LicenseObject::activate_with_type(const std::string &activation_code, LicenseActivationType activation_type)
{
    KLOG_PROFILE("License code: %s register type: %d.", activation_code.c_str(), activation_type);

    if (this->name_ != LICENSE_OBJECT_OS_NAME &&
        Glib::file_test(KYLINSEC_OS_FILE, Glib::FILE_TEST_EXISTS))
    {
        auto os_object = Daemon::Daemon::get_instance()->get_license_object(LICENSE_OBJECT_OS_NAME);
        auto license_info = os_object->get_license_info();
        // 未激活麒麟信安操作系统不允许激活增值业务
        if (license_info.licensed != LicenseActivationStatus::LAS_ACTIVATED)
        {
            return LicenseErrorCode::LEC_UNACTIVE_OS_FORBID_ACTIVE_VA;
        }
        // 订阅模式下激活已过期不允许激活增值业务
        if (license_info.expired_time < time(NULL) && license_info.update_mode == LicenseUpdateMode::LICENSE_UPDATE_MODE_SUBSCRIBE)
        {
            return LicenseErrorCode::LEC_OS_LICENSE_EXPIRY_FORBID_ACTIVE_VA;
        }
    }

    RETURN_VAL_IF_ERROR(this->check_activation_code(activation_code, activation_type));

    auto license_object_group = this->configuration_->get_license_object_group(this->name_);
    RETURN_VAL_IF_FALSE(license_object_group, LicenseErrorCode::LEC_COMMON_NULL_POINTER);

    license_object_group->activation_code = activation_code;
    license_object_group->activation_type = activation_type;
    license_object_group->activation_time = Base::TimeUtils::timestamp2str(time(NULL));
    license_object_group->once_active = true;

    if (this->configuration_->set_license_object_group(license_object_group))
    {
        // 记录注册成功的日志信息
        KLOG_INFO("Registration successful. Registeration Type: %d, machine code: %s, license code: %s.",
                  activation_type,
                  this->get_machine_code().c_str(),
                  activation_code.c_str());

        this->activation_success_.emit();
    }
    else
    {
        return LicenseErrorCode::LEC_SYNC_TO_FILE_FAILED;
    }
    return LicenseErrorCode::LEC_SUCCESS;
}

// 对比本地机器码和授权码解密后的机器码是否相等
bool LicenseObject::compare_machine_code(const std::string &local_str, const std::string &activation_str)
{
    std::string local_machine_code;
    std::string activation_machine_code;

    RETURN_VAL_IF_FALSE(Base::LicenseUtils::unformat_machine_code(local_str, local_machine_code) == LicenseErrorCode::LEC_SUCCESS, false);
    RETURN_VAL_IF_FALSE(Base::LicenseUtils::unformat_machine_code(activation_str, activation_machine_code) == LicenseErrorCode::LEC_SUCCESS, false);

    // 后7个字节为实际的机器码
    if (memcmp(local_machine_code.data() + 1, activation_machine_code.data() + 1, MACHINE_CODE_LENGTH - 1) != 0)
    {
        KLOG_WARNING("The 7 bytes after machine code is dismatch.");
        return false;
    }

    // 激活码解密后第1个字节的前4位为安装类型，为了兼容以前的情况，老的安装类型不参与对比
    LicenseInstallType activation_install_type = LicenseInstallType(activation_machine_code[0] & 0xF);
    if (activation_install_type > LICENSE_INSTALL_TYPE_OLD_LAST &&
        activation_install_type < LICENSE_INSTALL_TYPE_LAST &&
        activation_install_type != LICENSE_INSTALL_TYPE_UNLIMIT &&
        activation_install_type != local_machine_code[0])
    {
        KLOG_WARNING("The install type is dismatch.");
        return false;
    }
    // 未知形态不应该比较成功，用户可能删除或修改了配置文件 （不再限定该条件，会导致旧的系统无法适配成功）
    // RETURN_VAL_IF_TRUE(local_machine_code[0] == LICENSE_INSTALL_TYPE_UNKNOWN, false);

    return true;
}

void LicenseObject::update_license_info()
{
    auto license_info = this->get_license_info();

    if (this->cached_license_info_ != license_info)
    {
        this->cached_license_info_ = license_info;
        this->license_info_changed_.emit();
    }
}

void LicenseObject::idle_check_license()
{
    if (!this->idle_timer_)
    {
        auto idle = Glib::MainContext::get_default()->signal_idle();
        this->idle_timer_ = idle.connect(sigc::mem_fun(this, &LicenseObject::timeout_check_license));
    }
}

void LicenseObject::activate_online_with_config()
{
    // 如果是未激活或者是已激活但过期，则向激活服务器重新注册
    do
    {
        auto license_object_group = this->configuration_->get_license_object_group(this->name_);
        BREAK_IF_FALSE(license_object_group);

        BREAK_IF_TRUE(license_object_group->auth_server.length() == 0);
        BREAK_IF_TRUE(license_object_group->service_uid.length() == 0);

        auto error_code = this->activate_online(license_object_group->auth_server, license_object_group->service_uid);
        if (error_code != LicenseErrorCode::LEC_SUCCESS)
        {
            auto error = LICENSE_ERROR2STR(error_code);
            KLOG_WARNING("%s", error.c_str());
        }
    } while (0);
}

LicenseExceptionStatus LicenseObject::get_exception_status(time_t keep_alive_deadline)
{
    LicenseExceptionStatus exception_status = LICENSE_EXCEPTION_STATUS_NONE;
    if (keep_alive_deadline > 0 && keep_alive_deadline < time(NULL))
    {
        exception_status = LicenseExceptionStatus::LICENSE_EXCEPTION_STATUS_SERVER_NOT_CONNECTED;
    }

    if (this->server_timestamp_ != 0 &&
        (this->server_timestamp_ > time(NULL) + CS_TIMESTAMP_MAX_DIFF ||
         time(NULL) > this->server_timestamp_ + CS_TIMESTAMP_MAX_DIFF))
    {
        exception_status = LicenseExceptionStatus::LICENSE_EXCEPTION_STATUS_CS_TIME_DIFF_TOO_MUCH;
    }

    return exception_status;
}

void LicenseObject::fill_license_info(LicenseInfo &license_info, const std::shared_ptr<Base::ActivationCodeFields> &activation_code_fileds)
{
    auto license_object_group = this->configuration_->get_license_object_group(this->name_);

    RETURN_IF_FALSE(license_object_group);

    // 未激活的系统拷贝当前机器的机器码，已激活的系统使用激活码中的机器码
    if (license_info.licensed == LicenseActivationStatus::LAS_UNACTIVATED)
    {
        // 这里要将开始时间转换为当天的0点
        auto first_boot_gdate = Base::TimeUtils::timestamp2gdate(license_info.first_boot_time);
        license_info.expired_time = Base::TimeUtils::gdate2timestamp(&first_boot_gdate) + MONTH_IN_SECONDS * 6 + DAY_IN_SECONDS - 1;
        license_info.machine_code = this->get_machine_code();
        /* 如果当前时间比开始时间还要小，说明用户手动调整了时间，这种情况在试用期是不允许的
        不过为了避免存在一些误差，这里做一个月的宽限期,该异常仅需在未激活状态提示*/
        if (license_info.first_boot_time > time(NULL) + 30 * 24 * 3600)
        {
            // 兼容旧版本，此枚举标识符已弃用
            license_info.service_status = LicenseServiceStatus::LICENSE_SERVICE_STATUS_INVALID;
            license_info.exception_status = LicenseExceptionStatus::LICENSE_EXCEPTION_STATUS_INVALID;
        }
    }
    else if (license_info.licensed == LicenseActivationStatus::LAS_ACTIVATED)
    {
        license_info.activation_code = license_object_group->activation_code;
        license_info.customer_code = activation_code_fileds->customer_info;
        // 由于过期日期是按天为单位的，所以时间戳设置为当天的最后一秒
        license_info.expired_time = activation_code_fileds->expiry_timestamp;
        // 激活方式以激活码中记录的优先，新的格式都会在激活码中记录。激活码中没记录再使用文件中的，因为文件中是通过用户行为添加的，不太可信
        if (activation_code_fileds->activation_type != LicenseActivationType::LICENSE_ACTIVATION_TYPE_UNKNOWN)
        {
            license_info.activation_type = activation_code_fileds->activation_type;
        }
        else
        {
            license_info.activation_type = license_object_group->activation_type;
        }
        license_info.activation_time = Base::TimeUtils::str2timestamp(license_object_group->activation_time.c_str());
        license_info.update_mode = activation_code_fileds->update_mode;
        license_info.service_mode = activation_code_fileds->service_mode;
        license_info.service_status = time(NULL) <= license_info.expired_time
                                          ? LicenseServiceStatus::LICENSE_SERVICE_STATUS_UNEXPIRED
                                          : LicenseServiceStatus::LICENSE_SERVICE_STATUS_EXPIRED;
    }
}

bool LicenseObject::deactive_license()
{
    auto license_info = this->get_license_info();

    // 未激活，不清除
    if (license_info.licensed == LicenseActivationStatus::LAS_UNACTIVATED)
    {
        return false;
    }
    else
    {
        return this->clear_license(license_info.activation_type == LicenseActivationType::LICENSE_ACTIVATION_TYPE_ONLINE);
    }
}

bool LicenseObject::clear_license(bool online_active)
{
    auto license_object_group = this->configuration_->get_license_object_group(this->name_);
    RETURN_VAL_IF_FALSE(license_object_group, false);

    if (online_active)
    {
        license_object_group->auth_server.clear();
        license_object_group->keep_alive_deadline = 0;
        license_object_group->service_uid.clear();

        if (this->heartbeat_timer_.connected())
        {
            this->heartbeat_timer_.disconnect();
        }
    }
    license_object_group->activation_code.clear();
    license_object_group->activation_time.clear();
    license_object_group->activation_type = LicenseActivationType::LICENSE_ACTIVATION_TYPE_UNKNOWN;
    license_object_group->certificate_expired_time.clear();

    return this->configuration_->set_license_object_group(license_object_group);
}

bool LicenseObject::on_heartbeat_connection()
{
#ifdef ENABLE_SERVER
    auto license_object_group = this->configuration_->get_license_object_group(this->name_);

    RETURN_VAL_IF_FALSE(license_object_group->auth_server.length() != 0, true);

    grpc::ClientContext context;
    ::HeartBeatReq req;
    ::HeartBeatRsp rsp;
    auto channel = grpc::CreateChannel(fmt::format("{0}:{1}", license_object_group->auth_server, DEFAULT_LISTEN_PORT), grpc::InsecureChannelCredentials());
    auto stub = License::NewStub(channel);

    gpr_timespec deadline;
    deadline.tv_sec = LICENSE_CLIENT_TIMEOUT_SEC;
    deadline.tv_nsec = 0;
    deadline.clock_type = GPR_TIMESPAN;

    context.set_deadline(deadline);
    req.set_machine_code(this->get_machine_code());
    req.set_service_uid(license_object_group->service_uid);
    auto status = stub->HeartBeat(&context, req, &rsp);

    if (status.ok() && rsp.error_code() == LicenseErrorCode::LEC_SUCCESS)
    {
        license_object_group->keep_alive_deadline = (rsp.keep_alive_time() == 0) ? 0 : rsp.timestamp() + rsp.keep_alive_time();
        this->server_timestamp_ = time_t(rsp.timestamp());
        this->configuration_->set_license_object_group(license_object_group);
        KLOG_DEBUG("get heartbeat!");
    }
    else if (status.ok() && rsp.error_code() != LicenseErrorCode::LEC_SUCCESS)
    {
        /* 在以下情况可能会在服务器上找不到激活信息：1) 人为清理；2) 将激活信息同步到数据库之前服务重启了
           因此在获取不到激活信息的情况下，重新请求激活，确保服务端存在激活记录，否则心跳无法正常返回导致客户端的激活码不可用 */
        if (rsp.error_code() == LicenseErrorCode::LEC_SERVER_ACTIVATION_RECORD_NOT_FOUND)
        {
            this->activate_online_with_config();
        }

        if (rsp.error_code() > 0)
        {
            KLOG_WARNING("%s", LICENSE_ERROR2STR(LicenseErrorCode(rsp.error_code())).c_str());
        }
    }
    else
    {
        this->server_timestamp_ = 0;
        KLOG_WARNING("Failed get heartbeat, status code: %d.", int32_t(status.error_code()));
    }
#endif
    return true;
}

void LicenseObject::on_license_group_changed(const std::string &group_name)
{
    auto license_object_group_name = fmt::format("{0} {1}", LICENSE_OBJECT_KEYFILE_GROUP, this->get_name());
    RETURN_IF_TRUE(license_object_group_name != group_name);
    this->idle_check_license();
}

void LicenseObject::on_license_info_changed()
{
    // 如果是在线激活且已过期，则重新激活
    auto license_info = this->get_license_info();

    if (license_info.activation_type == LicenseActivationType::LICENSE_ACTIVATION_TYPE_ONLINE &&
        (license_info.service_status != LicenseServiceStatus::LICENSE_SERVICE_STATUS_UNEXPIRED))
    {
        this->activate_online_with_config();
    }
}

void LicenseObject::on_activation_success()
{
    auto license_object_group = this->configuration_->get_license_object_group(this->name_);

    if (license_object_group->activation_type == LicenseActivationType::LICENSE_ACTIVATION_TYPE_OFFLINE ||
        license_object_group->activation_type == LicenseActivationType::LICENSE_ACTIVATION_TYPE_MANNUAL)
    {
        license_object_group->auth_server.clear();
        this->configuration_->set_license_object_group(license_object_group);

        if (this->heartbeat_timer_)
        {
            KLOG_INFO("the heartbeat timer disconnected!");
            this->heartbeat_timer_.disconnect();
        }
    }

    if (license_object_group->activation_type == LicenseActivationType::LICENSE_ACTIVATION_TYPE_ONLINE &&
        !this->heartbeat_timer_.connected())
    {
        auto timeout = Glib::MainContext::get_default()->signal_timeout();
        this->heartbeat_timer_ = timeout.connect_seconds(sigc::mem_fun(this, &LicenseObject::on_heartbeat_connection), HEART_BEAT_INTERVAL);
    }
}

}  // namespace Daemon
