/**
 * @file          /ks-license/src/daemon/license-object.h
 * @brief
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */
#pragma once

#include "src/daemon/license-configuration.h"
#include "src/daemon/ukey-device.h"

namespace Daemon
{
struct LicenseInfo
{
    LicenseInfo();
    // 激活状态
    LicenseActivationStatus licensed;
    // 服务状态
    LicenseServiceStatus service_status;
    // 激活码
    std::string activation_code;
    // 机器码
    std::string machine_code;
    // 客户代码
    std::string customer_code;
    // 过期时间
    time_t expired_time;
    // 激活方式，如果为LICENSE_ACTIVATION_TYPE_UNKNOWN，则激活时间字段不可用
    LicenseActivationType activation_type;
    // 第一次使用系统的时间
    time_t first_boot_time;
    // 激活时间
    time_t activation_time;
    // 激活服务器地址
    std::string auth_server;
    // 服务号
    std::string service_uid;
    // 更新模式
    LicenseUpdateMode update_mode;
    // 服务模式
    LicenseServiceMode service_mode;
    // 异常状态
    LicenseExceptionStatus exception_status;

    bool operator==(const LicenseInfo &other)
    {
        return (this->licensed == other.licensed &&
                this->service_status == other.service_status &&
                this->activation_code == other.activation_code &&
                this->expired_time == other.expired_time &&
                this->activation_type == other.activation_type &&
                this->first_boot_time == other.first_boot_time &&
                this->activation_time == other.activation_time &&
                this->auth_server == other.auth_server);
    }

    bool operator!=(const LicenseInfo &other)
    {
        return !(this->operator==(other));
    }
};

class LicenseObject
{
public:
    LicenseObject(const std::string &name);
    virtual ~LicenseObject();

    bool timeout_check_license();
    bool timing_check_license();

    // 获取授权对象名称
    std::string get_name() { return this->name_; };

    // 获取授权信息
    LicenseInfo get_license_info();

    // 获取机器码
    std::string get_machine_code();

    // 校验激活码是否合法
    LicenseErrorCode check_activation_code(const std::string &activation_code, LicenseActivationType activation_type);

    std::shared_ptr<Base::ActivationCodeFields> check_activation_code(const std::string &activation_code,
                                                                      LicenseErrorCode &error_code);

    // 通过激活码激活
    LicenseErrorCode activate_by_activation_code(const std::string &activation_code);

    // 在线激活
    LicenseErrorCode activate_online(const std::string &host, const std::string &service_uid);

    // 通过UsbKey注册
    LicenseErrorCode activate_by_usb_key(std::shared_ptr<UKeyDevice> &used_ukey_device);

    // 设定激活方式激活
    LicenseErrorCode activate_with_type(const std::string &activation_code, LicenseActivationType activation_type);

    // 获取密钥
    std::string get_key() const { return this->key_; };

    bool deactive_license();

    sigc::signal<void> &signal_activation_success() { return this->activation_success_; };
    sigc::signal<void> &signal_license_info_changed() { return this->license_info_changed_; };

private:
    bool compare_machine_code(const std::string &local_str, const std::string &license_str);

    // 更新授权信息
    void update_license_info();

    // 空闲时检查license状态变化情况
    void idle_check_license();

    void activate_online_with_config();

    bool clear_license(bool active_online);

    LicenseExceptionStatus get_exception_status(time_t keep_alive_deadline);
    // 判断激活状态，并通过激活状态填充 服务模式、更新模式、异常状态、服务状态
    void fill_license_info(LicenseInfo &license_info, const std::shared_ptr<Base::ActivationCodeFields> &activation_code_fileds);

    // 发送心跳连接
    bool on_heartbeat_connection();
    void on_license_group_changed(const std::string &group_name);
    void on_license_info_changed();
    void on_activation_success();

protected:
    // 授权对象的名称
    std::string name_;
    std::shared_ptr<LicenseConfiguration> configuration_;

private:
    // 密钥
    std::string key_;

    sigc::signal<void> activation_success_;
    sigc::signal<void> license_info_changed_;

    // 定时器，用于定时检查授权情况
    sigc::connection timing_timer_;
    // 授权已过期时检查授权状态
    sigc::connection expired_timer_;
    // 心跳连接定时器
    sigc::connection heartbeat_timer_;
    // 空闲时检查授权状态
    sigc::connection idle_timer_;

    LicenseInfo cached_license_info_;

    std::vector<std::string> mac_addrs_;

    // 服务器时间戳，用于检测客户端和服务端时间是否同步
    time_t server_timestamp_;
};

typedef std::vector<std::shared_ptr<LicenseObject>> LicenseObjectVec;
}  // namespace Daemon
