/**
 * @file          /kylin-license/src/daemon/ukey-device.cpp
 * @brief
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */

#include "src/daemon/ukey-device.h"

#include <fcntl.h>
#include <glib/gi18n.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "lib/base/base.h"

namespace Daemon
{
#define RY3_LOW_LEVEL_PLACEHOLDER_1 'K'
#define RY3_LOW_LEVEL_PLACEHOLDER_2 'Y'

#define ROCKEY_NOTIFY_UPDATE "1"
#define ROCKEY_FILE_PERMISSION 0644

enum PUKeyAddrOffset
{
    // 服务号
    PUKEY_ADDR_OFFSET_UID = 7168,
    // 硬件序列号
    PUKEY_ADDR_OFFSET_HARD_ID = 7200,
    // 客户信息
    PUKEY_ADDR_OFFSET_CUSTOMER_INFO = 7216,
    // 激活对象
    PUKEY_ADDR_OFFSET_OBJECT_NAME = 7240,
    // 质保期
    PUKEY_ADDR_OFFSET_EXPIRY_DATE = 7256,
    // 安装类型
    PUKEY_ADDR_OFFSET_INSTALL_TYPE = 7264,
    // 更新模式
    PUKEY_ADDR_OFFSET_UPDATE_MODE = 7265,
    // 人工服务模式
    PUKEY_ADDR_OFFSET_ARTIFICIAL_SERVICE_MODE = 7266,
    // 总激活点数
    PUKEY_ADDR_OFFSET_SUM_POINTS = 7267,
    // 剩余激活点数
    PUKEY_ADDR_OFFSET_REMAINING_POINTS = 7271,
    PUKEY_ADDR_OFFSET_CUR_LAST = 7275,
    PUKEY_ADDR_OFFSET_LAST = 8192,
};

// 服务号
#define RY3_UID_BYTES (PUKeyAddrOffset::PUKEY_ADDR_OFFSET_HARD_ID - PUKeyAddrOffset::PUKEY_ADDR_OFFSET_UID)
// 硬件序列号偏移(16字节)
#define RY3_HARDINFO_BYTES (PUKeyAddrOffset::PUKEY_ADDR_OFFSET_CUSTOMER_INFO - PUKeyAddrOffset::PUKEY_ADDR_OFFSET_HARD_ID)
// 客户信息
#define RY3_CUSTOMER_INFO_BYTES (PUKeyAddrOffset::PUKEY_ADDR_OFFSET_OBJECT_NAME - PUKeyAddrOffset::PUKEY_ADDR_OFFSET_CUSTOMER_INFO)
// 激活对象
#define RY3_OBJECT_NAME_BYTES (PUKeyAddrOffset::PUKEY_ADDR_OFFSET_EXPIRY_DATE - PUKeyAddrOffset::PUKEY_ADDR_OFFSET_OBJECT_NAME)
// 质保期
#define RY3_EXPIRY_DATE_BYTES (PUKeyAddrOffset::PUKEY_ADDR_OFFSET_INSTALL_TYPE - PUKeyAddrOffset::PUKEY_ADDR_OFFSET_EXPIRY_DATE)
// 安装类型
#define RY3_INSTALL_TYPE_BYTES (PUKeyAddrOffset::PUKEY_ADDR_OFFSET_UPDATE_MODE - PUKeyAddrOffset::PUKEY_ADDR_OFFSET_INSTALL_TYPE)
// 更新模式
#define RY3_UDPATE_MODE_BYTES (PUKeyAddrOffset::PUKEY_ADDR_OFFSET_ARTIFICIAL_SERVICE_MODE - PUKeyAddrOffset::PUKEY_ADDR_OFFSET_UPDATE_MODE)
// 人工服务模式
#define RY3_ARTIFICIAL_SERVICE_MODE_BYTES (PUKeyAddrOffset::PUKEY_ADDR_OFFSET_SUM_POINTS - PUKeyAddrOffset::PUKEY_ADDR_OFFSET_ARTIFICIAL_SERVICE_MODE)
// 总激活点数
#define RY3_SUM_POINTS_BYTES (PUKeyAddrOffset::PUKEY_ADDR_OFFSET_REMAINING_POINTS - PUKeyAddrOffset::PUKEY_ADDR_OFFSET_SUM_POINTS)
// 剩余激活点数
#define RY3_REMAINING_POINTS_BYTES (PUKeyAddrOffset::PUKEY_ADDR_OFFSET_CUR_LAST - PUKeyAddrOffset::PUKEY_ADDR_OFFSET_REMAINING_POINTS)

UKeyLowAddress::UKeyLowAddress() : item_count(0)
{
}

UKeyDevice::UKeyDevice(RY_HANDLE handle) : handle_(handle)
{
}

UKeyDevice::~UKeyDevice()
{
    if (this->handle_)
    {
#if (defined(__x86_64__) || defined(__aarch64__)) && defined(ENABLE_ROCKEY)
        RY3_Close(this->handle_, 1);
#endif
        this->handle_ = NULL;
    }
}

bool UKeyDevice::init()
{
    std::string error;
    if (!this->load_ukey(error))
    {
        KLOG_WARNING("%s", error.c_str());
        return false;
    }
    return true;
}

bool UKeyDevice::is_expired()
{
    auto now_gdate = Base::TimeUtils::get_current_gdate();
    RETURN_VAL_IF_TRUE(g_date_compare(&now_gdate, &this->high_address_area_.expiry_date) > 0, true);
    return false;
}

bool UKeyDevice::exist_machine_code(const std::string &machine_code)
{
    auto iter = this->machine_codes_.find(machine_code);
    if (iter == this->machine_codes_.end())
    {
        return false;
    }
    return true;
}

LicenseErrorCode UKeyDevice::generate_activation_code(const std::string &machine_code_str,
                                                      std::string &out_license_code)
{
#if (defined(__x86_64__) || defined(__aarch64__)) && defined(ENABLE_ROCKEY)
    UKeyLowAddressItem new_item;
    unsigned long retcode = 0;
    std::string machine_code;
    unsigned char expiry_date_buf[OLD_EXPIRY_DATE_LENGTH];
    unsigned char in_buf[LICENSE_CODE_LENGTH + KEY_LEN] = {0};
    std::string out_buf = std::string(LICENSE_CODE_LENGTH, '\0');
    int out_buf_size = 0;

    RETURN_VAL_IF_ERROR(Base::LicenseUtils::unformat_machine_code(machine_code_str, machine_code));
    auto key = Base::LicenseUtils::generate_key(this->high_address_area_.activation_object_name);

    memcpy(in_buf, machine_code.data(), MACHINE_CODE_LENGTH);

    if (this->high_address_area_.customer_info.length() > 0)
    {
        // 旧的激活码格式：机器码+质保期+客户信息
        RETURN_VAL_IF_ERROR(Base::TimeUtils::gdate2arr(&(this->high_address_area_.expiry_date), expiry_date_buf));
        memcpy(in_buf + MACHINE_CODE_LENGTH, expiry_date_buf, OLD_EXPIRY_DATE_LENGTH);
        memcpy(in_buf + MACHINE_CODE_LENGTH + OLD_EXPIRY_DATE_LENGTH, this->high_address_area_.customer_info.c_str(), CUSTOMER_CODE_LENGTH);
    }
    else
    {
        // 新的激活码格式：机器码+质保期+更新模式+服务模式
        // 第一个字节最高位标记为1表示新的格式
        in_buf[0] |= ACTIVATION_CODE_VERSION_MASK;
        // 由于过期日期是按天为单位的，所以时间戳设置为当天的最后一秒
        uint32_t expiry_timestamp = uint32_t(Base::TimeUtils::gdate2timestamp(&this->high_address_area_.expiry_date) + DAY_IN_SECONDS - 1);
        memcpy(in_buf + MACHINE_CODE_LENGTH, &expiry_timestamp, EXPIRY_DATE_LENGTH);

        in_buf[MACHINE_CODE_LENGTH + EXPIRY_DATE_LENGTH] |= (this->high_address_area_.update_mode << LICENSE_UPDATE_MODE_BIT_POS);
        in_buf[MACHINE_CODE_LENGTH + EXPIRY_DATE_LENGTH] |= (this->high_address_area_.service_mode << LICENSE_SERVICE_MODE_BIT_POS);
        in_buf[MACHINE_CODE_LENGTH + EXPIRY_DATE_LENGTH] |= (LICENSE_ACTIVATION_TYPE_OFFLINE << LICENSE_ACTIVATION_TYPE_BIT_POS);
    }
    memcpy(in_buf + LICENSE_CODE_LENGTH, key.c_str(), KEY_LEN);

    out_buf_size = out_buf.length();
    retcode = RY3_ExecuteFile(this->handle_, 0x0001, in_buf, sizeof(in_buf), (unsigned char *)out_buf.data(), &out_buf_size);
    if (retcode != RY3_SUCCESS)
    {
        return LicenseErrorCode::LEC_C51_GENERATE_LICENSE_FAILED;
    }

    // 判断之前是否已经注册过
    if (this->exist_machine_code(machine_code_str))
    {
        KLOG_DEBUG("The machine is already registered.");
    }
    else
    {
        new_item.machine_code = machine_code;
        new_item.placeholder[0] = RY3_LOW_LEVEL_PLACEHOLDER_1;
        new_item.placeholder[1] = RY3_LOW_LEVEL_PLACEHOLDER_2;
        RETURN_VAL_IF_ERROR(this->write_low_address_item(new_item));
        this->machine_codes_.insert(machine_code_str);
    }

    out_license_code = Base::LicenseUtils::format_activation_code(out_buf);

    this->print_ukey();
    return LicenseErrorCode::LEC_SUCCESS;
#else
    return LicenseErrorCode::LEC_UKEY_UNSUPPORTED_ARCH;
#endif
}

void UKeyDevice::print_ukey()
{
    KLOG_INFO("Handle:                      %p", this->handle_);
    KLOG_INFO("HardID(1-2):                 %s-%s", this->hard_id_.c_str(), this->high_address_area_.hard_id.c_str());
    KLOG_INFO("CustomerInfo:                %s", this->high_address_area_.customer_info.c_str());
    KLOG_INFO("ActivationObjectName:        %s", this->high_address_area_.activation_object_name.c_str());
    KLOG_INFO("ExpredDate:                  %04d-%02d-%02d",
              this->high_address_area_.expiry_date.year,
              this->high_address_area_.expiry_date.month,
              this->high_address_area_.expiry_date.day);
    KLOG_INFO("InstallType:                 %d", this->high_address_area_.install_type);
    KLOG_INFO("UpdateMode:                  %d", this->high_address_area_.update_mode);
    KLOG_INFO("ServiceMode:                 %d", this->high_address_area_.service_mode);
    KLOG_INFO("SumPoints:                   %d", this->high_address_area_.sum_points);
    KLOG_INFO("RemainingPoints:             %d", this->high_address_area_.remaining_points);
    KLOG_INFO("ActivationMachineCodes:");

    for (gint32 j = 0; j < this->low_address_area_.item_count; ++j)
    {
        auto machine_code_str = Base::LicenseUtils::format_machine_code(this->low_address_area_.item[j].machine_code);
        KLOG_INFO("    %s", machine_code_str.c_str());
    }
}

bool UKeyDevice::load_ukey(std::string &error)
{
    KLOG_PROFILE("");

#if (defined(__x86_64__) || defined(__aarch64__)) && defined(ENABLE_ROCKEY)
    char hard_id[RY3_HARDINFO_BYTES + 1] = {0};
    // 获取硬件序列号
    auto retcode = RY3_GetHardID(this->handle_, hard_id);
    KLOG_WARNING("HardID: %s.", hard_id);
    RETURN_VAL_AND_ERR_IF_TRUE(retcode != RY3_SUCCESS, false, error, _("Failed to read hard id"));

    this->hard_id_ = hard_id;
    this->pincode_ = this->hard_id_ + PIN_SUFFIX;

    RETURN_VAL_IF_FALSE(this->load_high_address_area(error), false);
    RETURN_VAL_AND_ERR_IF_TRUE(this->hard_id_ != this->high_address_area_.hard_id, false, error, _("The hard id is unmatch"));

    // 计算过期也需要加载进来，便于给到用户正确的错误提示
    // auto now_gdate = TimeUtils::get_current_gdate();
    // RETURN_VAL_AND_ERR_IF_TRUE(g_date_compare(&now_gdate, &this->high_address_area_.expire_time) > 0,
    //                            false,
    //                            error,
    //                            _("The USBKey device is expired, hard id: {0}"),
    //                            this->hard_id_);

    this->low_address_area_.item_count = this->get_used_points();
    RETURN_VAL_IF_FALSE(this->load_low_address_area(0, this->low_address_area_.item_count, error), false);
    this->print_ukey();
#endif
    return true;
}

bool UKeyDevice::load_high_address_area(std::string &error)
{
#if (defined(__x86_64__) || defined(__aarch64__)) && defined(ENABLE_ROCKEY)
    unsigned long retcode = 0;
    char customer_info[RY3_CUSTOMER_INFO_BYTES];
    char hard_id[RY3_HARDINFO_BYTES + 1];
    char buffer[BUFSIZ];

#define READ_FIELD(pos, var, size)                                                                          \
    {                                                                                                       \
        auto retcode = RY3_Read(this->handle_, pos, (unsigned char *)(var), size);                          \
        RETURN_VAL_AND_ERR_IF_FALSE(retcode == RY3_SUCCESS, false, error, _("Failed to read USBKey data")); \
    }

    // 服务号
    READ_FIELD(PUKeyAddrOffset::PUKEY_ADDR_OFFSET_UID, buffer, RY3_UID_BYTES);
    this->high_address_area_.uid = std::string(buffer, RY3_UID_BYTES);

    // 硬件序列号
    READ_FIELD(PUKeyAddrOffset::PUKEY_ADDR_OFFSET_HARD_ID, buffer, RY3_HARDINFO_BYTES);
    this->high_address_area_.hard_id = std::string(buffer, RY3_HARDINFO_BYTES);

    // 客户信息
    READ_FIELD(PUKeyAddrOffset::PUKEY_ADDR_OFFSET_CUSTOMER_INFO, buffer, RY3_CUSTOMER_INFO_BYTES);
    this->high_address_area_.customer_info = std::string(buffer);

    // 激活对象
    READ_FIELD(PUKeyAddrOffset::PUKEY_ADDR_OFFSET_OBJECT_NAME, buffer, RY3_OBJECT_NAME_BYTES);
    this->high_address_area_.activation_object_name = std::string(buffer);

    // 质保期
    READ_FIELD(PUKeyAddrOffset::PUKEY_ADDR_OFFSET_EXPIRY_DATE, &(this->high_address_area_.expiry_date), RY3_EXPIRY_DATE_BYTES);
    // 安装类型
    READ_FIELD(PUKeyAddrOffset::PUKEY_ADDR_OFFSET_INSTALL_TYPE, &(this->high_address_area_.install_type), RY3_INSTALL_TYPE_BYTES);
    // 更新模式
    READ_FIELD(PUKeyAddrOffset::PUKEY_ADDR_OFFSET_UPDATE_MODE, &(this->high_address_area_.update_mode), RY3_UDPATE_MODE_BYTES);
    // 人工服务模式
    READ_FIELD(PUKeyAddrOffset::PUKEY_ADDR_OFFSET_ARTIFICIAL_SERVICE_MODE, &(this->high_address_area_.service_mode), RY3_ARTIFICIAL_SERVICE_MODE_BYTES);
    // 总激活点数
    READ_FIELD(PUKeyAddrOffset::PUKEY_ADDR_OFFSET_SUM_POINTS, &(this->high_address_area_.sum_points), RY3_SUM_POINTS_BYTES);
    // 剩余激活点数
    READ_FIELD(PUKeyAddrOffset::PUKEY_ADDR_OFFSET_REMAINING_POINTS, &(this->high_address_area_.remaining_points), RY3_REMAINING_POINTS_BYTES);

#undef READ_FIELD

#endif
    return true;
}

bool UKeyDevice::load_low_address_area(int32_t start, int32_t len, std::string &error)
{
    KLOG_PROFILE("start: %d, len: %d", start, len);

    auto buffer = new char[UKEY_MACHINE_CODE_ITEM_LEN * len + 1];
    bool retval = false;

    do
    {
        if (len == 0)
        {
            retval = true;
            break;
        }
        BREAK_AND_ERR_IF_TRUE(len < 0, error, _("The remaining points is greater than activation points"));

#if (defined(__x86_64__) || defined(__aarch64__)) && defined(ENABLE_ROCKEY)

        KLOG_DEBUG("Read low level info. start index: %d, fiiled number: %d.", start, len);

        unsigned long retcode = RY3_Read(this->handle_,
                                         UKEY_MACHINE_CODE_ITEM_LEN * start,
                                         (unsigned char *)buffer,
                                         UKEY_MACHINE_CODE_ITEM_LEN * len);
        BREAK_AND_ERR_IF_TRUE(retcode != RY3_SUCCESS, error, _("Failed to read machine codes from USBKey"));

#endif
        for (int32_t i = start; i < start + (int)len; ++i)
        {
            this->low_address_area_.item[i].machine_code = std::string(buffer + (i - start) * UKEY_MACHINE_CODE_ITEM_LEN, MACHINE_CODE_LENGTH);
            memcpy(this->low_address_area_.item[i].placeholder, buffer + (i - start) * UKEY_MACHINE_CODE_ITEM_LEN + MACHINE_CODE_LENGTH, MACHINE_CODE_CHECK_BIT_LEN);

            if (this->low_address_area_.item[i].placeholder[0] != RY3_LOW_LEVEL_PLACEHOLDER_1 ||
                this->low_address_area_.item[i].placeholder[1] != RY3_LOW_LEVEL_PLACEHOLDER_2)
            {
                error = _("The machine code check bit recorded in USBKey is wrong");
                break;
            }
            auto machine_code_str = Base::LicenseUtils::format_machine_code(this->low_address_area_.item[i].machine_code);
            this->machine_codes_.insert(machine_code_str);
        }
        retval = true;
    } while (0);

    delete[] buffer;

    return retval;
}

LicenseErrorCode UKeyDevice::write_low_address_item(UKeyLowAddressItem &low_address_item)
{
    KLOG_PROFILE("");

#if (defined(__x86_64__) || defined(__aarch64__)) && defined(ENABLE_ROCKEY)
    unsigned long retcode = 0;
    int remain_count = 0;

    RETURN_VAL_IF_TRUE(this->low_address_area_.item_count > UKEY_MAX_AUTHORIZED_NUM,
                       LicenseErrorCode::LEC_UKEY_ITEM_EXCEED_LIMIT);
    RETURN_VAL_IF_TRUE(this->high_address_area_.remaining_points <= 0, LicenseErrorCode::LEC_UKEY_NO_REMAINING_POINTS);

    // 写高级存储区之前需要校验PIN码
    retcode = RY3_VerifyDevPin(this->handle_, (char *)this->pincode_.c_str(), &remain_count);
    RETURN_VAL_IF_TRUE(retcode != RY3_SUCCESS, LicenseErrorCode::LEC_UKEY_PIN_FAILED);

    // 写入机器码到ukey设备
    unsigned char machine_code_item[UKEY_MACHINE_CODE_ITEM_LEN + 1] = {0};
    memcpy(machine_code_item, low_address_item.machine_code.data(), MACHINE_CODE_LENGTH);
    memcpy(machine_code_item + MACHINE_CODE_LENGTH, low_address_item.placeholder, MACHINE_CODE_CHECK_BIT_LEN);
    retcode = RY3_Write(this->handle_,
                        this->low_address_area_.item_count * UKEY_MACHINE_CODE_ITEM_LEN,
                        (unsigned char *)machine_code_item,
                        UKEY_MACHINE_CODE_ITEM_LEN);
    RETURN_VAL_IF_TRUE(retcode != RY3_SUCCESS, LicenseErrorCode::LEC_UKEY_WRITE_MACHINE_CODE_FAILED);

    // 更新内存数据
    --this->high_address_area_.remaining_points;
    retcode = RY3_VendorWrite(this->handle_,
                              (int)PUKeyAddrOffset::PUKEY_ADDR_OFFSET_REMAINING_POINTS,
                              (unsigned char *)(&this->high_address_area_.remaining_points),
                              RY3_REMAINING_POINTS_BYTES);
    // 写入失败需要对点数进行还原
    if (retcode != RY3_SUCCESS)
    {
        ++this->high_address_area_.remaining_points;
        return LicenseErrorCode::LEC_UKEY_WRITE_REMAINING_POINTS_FAILED;
    }

    this->low_address_area_.item[this->low_address_area_.item_count++] = low_address_item;

#endif
    return LicenseErrorCode::LEC_SUCCESS;
}

}  // namespace Daemon
