/**
 * @file          /kylin-license/src/daemon/ukey-device.h
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */

#pragma once

#include <set>
#include "lib/base/base.h"

#if (defined(__x86_64__) || defined(__aarch64__)) && defined(ENABLE_ROCKEY)
#include "RY3_API.h"
#else
typedef void *RY_HANDLE;
#endif

namespace Daemon
{
// 机器码校验位长度
#define MACHINE_CODE_CHECK_BIT_LEN 2
// Ukey存储机器码使用的长度 (机器码长度+校验位长度)
#define UKEY_MACHINE_CODE_ITEM_LEN (MACHINE_CODE_LENGTH + MACHINE_CODE_CHECK_BIT_LEN)

// UKey低级存储区每项数据结构
struct UKeyLowAddressItem
{
    std::string machine_code;
    //两字节占位符,作为校验码
    unsigned char placeholder[MACHINE_CODE_CHECK_BIT_LEN];
};

// ukey 低级存储区数据结构
struct UKeyLowAddress
{
    UKeyLowAddress();
    UKeyLowAddressItem item[UKEY_MAX_AUTHORIZED_NUM];
    gint32 item_count;
};

// ukey 高级存储区数据结构
struct UKeyHighAddress
{
    UKeyHighAddress()
    {
        g_date_clear(&this->expiry_date, 1);
    };
    // 服务号
    std::string uid;
    // 硬件序列号
    std::string hard_id;
    // 客户信息
    std::string customer_info;
    // 激活对象
    std::string activation_object_name;
    // 质保期
    GDate expiry_date;
    // 安装类型
    uint8_t install_type;
    // 更新模式
    uint8_t update_mode;
    // 人工服务模式(5x8或7x24)
    uint8_t service_mode;
    // 总激活点数
    int32_t sum_points;
    // 剩余激活点数
    int32_t remaining_points;
};

class UKeyDevice
{
public:
    UKeyDevice(RY_HANDLE handle);
    virtual ~UKeyDevice();

    bool init();

    // 获取硬件序列号
    std::string get_hard_id() { return this->hard_id_; };

    // 获取高级存储区
    const UKeyHighAddress &get_high_address_area() { return this->high_address_area_; }

    // 已使用的点数
    int32_t get_used_points() { return this->high_address_area_.sum_points - this->high_address_area_.remaining_points; }

    // UKEY是否已经过期
    bool is_expired();

    // 判断机器码是否已经授权
    bool exist_machine_code(const std::string &machine_code);

    LicenseErrorCode generate_activation_code(const std::string &machine_code,
                                              std::string &out_license_code);

    virtual void print_ukey();

private:
    // 加载USBKey设备中的数据
    bool load_ukey(std::string &error);
    // 获取高级存储区数据
    bool load_high_address_area(std::string &error);
    // 获取低级存储区数据(UsbKey中存放的已授权信息)
    bool load_low_address_area(int32_t start, int32_t len, std::string &error);
    // 写入低级存储区
    LicenseErrorCode write_low_address_item(UKeyLowAddressItem &low_address_item);

private:
    // 设备句柄
    RY_HANDLE handle_;
    // 硬件序列号
    std::string hard_id_;
    // 开发商口令
    std::string pincode_;
    // 高级存储区数据
    UKeyHighAddress high_address_area_;
    // 低级存储区数据
    UKeyLowAddress low_address_area_;
    // 已授权的机器码
    std::set<std::string> machine_codes_;
};

}  // namespace Daemon
