/**
 * @file          /kylin-license/src/daemon/ukey-manager.cpp
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */

#include "src/daemon/ukey-manager.h"
#include <gudev/gudev.h>

#if (defined(__x86_64__) || defined(__aarch64__)) && defined(ENABLE_ROCKEY)
#include "RY3_API.h"
#endif

namespace Daemon
{
#define ROCKEY_SUBSYSTEM "usb"
#define ROCKEY_PROP_PRODUCT_KEY "product"
#define ROCKEY_PROP_PRODUCT_VALUE "ROCKEY3"

UKeyManager::UKeyManager() : udev_client_(NULL),
                             udev_client_handler_(0)
{
}

UKeyManager::~UKeyManager()
{
    if (this->udev_client_handler_)
    {
        g_signal_handler_disconnect(this->udev_client_, this->udev_client_handler_);
    }

    if (this->udev_client_)
    {
        g_object_unref(this->udev_client_);
    }
}

UKeyManager *UKeyManager::instance_ = NULL;
void UKeyManager::global_init()
{
    instance_ = new UKeyManager();
    instance_->init();
}

std::shared_ptr<UKeyDevice> UKeyManager::get_ukey(const std::string &activation_object_name)
{
    for (auto iter = this->ukeys_.begin(); iter != this->ukeys_.end(); ++iter)
    {
        auto ukey_device = iter->second;
        auto &high_address_area = ukey_device->get_high_address_area();
        if (high_address_area.activation_object_name == activation_object_name &&
            high_address_area.remaining_points > 0 &&
            !ukey_device->is_expired())
        {
            return ukey_device;
        }
    }
    return std::shared_ptr<UKeyDevice>();
}

std::shared_ptr<UKeyDevice> UKeyManager::get_ukey(const std::string &activation_object_name,
                                                  const std::string &machine_code_str)
{
    std::string machine_code;

    RETURN_VAL_IF_FALSE(Base::LicenseUtils::unformat_machine_code(machine_code_str, machine_code) == LicenseErrorCode::LEC_SUCCESS,
                        std::shared_ptr<UKeyDevice>());

    auto ukey_device = this->get_ukey_by_machine_code(machine_code_str);

    // 如果该机器没有注册过，或者注册过但是已经过期，则找一个可用的ukey
    if (!ukey_device || ukey_device->is_expired())
    {
        return this->get_ukey(activation_object_name);
    }

    return ukey_device;
}

void UKeyManager::init()
{
    const gchar *subsystem[2] = {ROCKEY_SUBSYSTEM, NULL};
    this->udev_client_ = g_udev_client_new(subsystem);
    this->udev_client_handler_ = g_signal_connect(this->udev_client_, "uevent", G_CALLBACK(UKeyManager::on_uevent_changed), this);
    this->load_physical_ukeys();
}

void UKeyManager::load_physical_ukeys()
{
    KLOG_PROFILE("");

#if (defined(__x86_64__) || defined(__aarch64__)) && defined(ENABLE_ROCKEY)

    unsigned long retcode = 0;
    int32_t ukey_num = 0;

    this->ukeys_.clear();

    // 查找设备
    retcode = RY3_Find(CLIENT_RY3_VENDOR_ID, &ukey_num);
    RETURN_IF_TRUE(retcode != RY3_SUCCESS);

    for (int32_t i = 1; i <= ukey_num; i++)
    {
        RY_HANDLE handle;
        // 打开设备
        retcode = RY3_Open(&handle, i);
        if (retcode != RY3_SUCCESS)
        {
            KLOG_WARNING("Failed to open %d-th USBKey device.", i);
            continue;
        }

        auto ukey_device = std::make_shared<UKeyDevice>(handle);
        if (ukey_device->init())
        {
            this->add_ukey(ukey_device);
        }
    }
#endif
}

void UKeyManager::add_ukey(std::shared_ptr<UKeyDevice> ukey_device)
{
    RETURN_IF_FALSE(ukey_device);

    if (this->ukeys_.find(ukey_device->get_hard_id()) == this->ukeys_.end())
    {
        this->ukeys_[ukey_device->get_hard_id()] = ukey_device;
    }
    else
    {
        KLOG_WARNING("The hard id '%s' is already exist", ukey_device->get_hard_id().c_str());
    }
}

std::shared_ptr<UKeyDevice> UKeyManager::get_ukey_by_machine_code(const std::string &machine_code)
{
    for (auto iter = this->ukeys_.begin(); iter != this->ukeys_.end(); ++iter)
    {
        auto ukey_device = iter->second;
        if (ukey_device->exist_machine_code(machine_code))
        {
            return ukey_device;
        }
    }
    return std::shared_ptr<UKeyDevice>();
}

void UKeyManager::on_uevent_changed(void *client, gchar *action, void *device, gpointer user_data)
{
    const gchar *name = NULL;
    const gchar *devtype = NULL;
    UKeyManager *self = (UKeyManager *)user_data;

    g_return_if_fail(device != NULL);

    name = g_udev_device_get_name((GUdevDevice *)device);
    devtype = g_udev_device_get_devtype((GUdevDevice *)device);

    KLOG_DEBUG("action: %s name: %s type: %s number: %s sysfs_path: %s.",
               action,
               name,
               devtype,
               g_udev_device_get_number((GUdevDevice *)device),
               g_udev_device_get_sysfs_path((GUdevDevice *)device));

    if (g_strcmp0(action, "add") != 0 && g_strcmp0(action, "remove") != 0)
    {
        return;
    }

    self->load_physical_ukeys();
}

}  // namespace Daemon
