#include "src/daemon/yum-repo-control.h"
#include <fcntl.h>
#include <stdarg.h>

#ifdef ENABLE_SERVER
#include <libsoup/soup.h>
#endif

namespace Daemon
{
#define DNF_VARS_DIR "/etc/dnf/vars"
#define DNF_VAR_OSVERSION "osversion"
#define DNF_VAR_OSVERSION_FILEPATH DNF_VARS_DIR "/" DNF_VAR_OSVERSION
#define REPO_NAME "ks-license-autogen"
#define CERTIFACATE_FILENAME "autogen.tar.gz"
#define CERTIFACATE_FILEPATH KL_INSTALL_DATADIR "/" CERTIFACATE_FILENAME
#define CERTIFACATE_FILE_CRT ".crt"
#define CERTIFACATE_FILE_KEY ".key"

#define LICENSE_FILE_PERMISSION 0644

#define REPO_KEY_ENABLE "enabled"
#define REPO_KEY_SSLVERIFY "sslverify"
#define REPO_KEY_SSLCLIENTCERT "sslclientcert"
#define REPO_KEY_SSLCLIENTKEY "sslclientkey"
#define REPO_ENABLE_VALUE 1
#define REPO_DISABLE_VALUE 0
#define REPO_SSLCLIENTCERT_FILE_NAME "/$osversion.crt"
#define REPO_SSLCLIENTKEY_FILE_NAME "/$osversion.key"

YumRepoControl::YumRepoControl(const std::string& repo_file_name) : repo_file_name_(repo_file_name),
                                                                    http_session_(NULL)
{
#ifdef ENABLE_SERVER
    this->http_session_ = soup_session_new_with_options(SOUP_SESSION_SSL_STRICT, FALSE, NULL);

    this->load_dnf_vars();
    this->dnf_vars_monitor_ = Base::FileUtils::make_monitor_file(DNF_VAR_OSVERSION_FILEPATH, sigc::mem_fun(this, &YumRepoControl::on_dnf_vars_changed));
    this->certificate_crt_monitor_ = Base::FileUtils::make_monitor_file(std::string(KL_INSTALL_DATADIR) + "/" + this->osversion_ + CERTIFACATE_FILE_CRT,
                                                                        sigc::mem_fun(this, &YumRepoControl::on_certificate_file_changed));
    this->certificate_key_monitor_ = Base::FileUtils::make_monitor_file(std::string(KL_INSTALL_DATADIR) + "/" + this->osversion_ + CERTIFACATE_FILE_KEY,
                                                                        sigc::mem_fun(this, &YumRepoControl::on_certificate_file_changed));

#endif
}
YumRepoControl::~YumRepoControl()
{
    if (this->http_session_)
    {
        g_object_unref(this->http_session_);
    }
}

void YumRepoControl::load_dnf_vars()
{
    try
    {
        std::string standard_error;
        this->osversion_ = Base::StrUtils::trim(Glib::file_get_contents(DNF_VAR_OSVERSION_FILEPATH));
        Glib::spawn_command_line_sync("arch", &this->arch_, &standard_error);
        this->arch_ = Base::StrUtils::trim(this->arch_);
    }
    catch (const Glib::Error& e)
    {
        KLOG_WARNING("%s", e.what().c_str());
    }
}

bool YumRepoControl::ping_certificate_server()
{
#ifdef ENABLE_SERVER
    auto msg = soup_message_new("POST", "https://mirrorlist.kylinsec.com.cn:9527/");
    auto status = soup_session_send_message(this->http_session_, msg);
    g_object_unref(msg);

    KLOG_DEBUG("Ping certificate server, return status code: %d.", status);
    return (status == SOUP_STATUS_OK);
#else
    return false;
#endif
}

bool YumRepoControl::download_certificate_file()
{
#ifdef ENABLE_SERVER
    RETURN_VAL_IF_TRUE(!this->ping_certificate_server(), false);

    this->enable_repo_groups(false);

    auto msg = soup_message_new("POST", "https://mirrorlist.kylinsec.com.cn:9527/obtain_dnf_certificate/");
    auto req_body = fmt::format("name={0}&version={1}&arch={2}&validity_period={3}", REPO_NAME, this->osversion_, this->arch_, DEFAULT_VALIDITY_PERIOD);
    KLOG_DEBUG("req_body = %s.", req_body.c_str());
    soup_message_set_request(msg, "application/x-www-form-urlencoded", SOUP_MEMORY_COPY, req_body.c_str(), req_body.size());
    auto status = soup_session_send_message(this->http_session_, msg);

    try
    {
        if (status == SOUP_STATUS_OK)
        {
            KLOG_DEBUG("Receive body length: %d.", msg->response_body->length);
            Glib::file_set_contents(CERTIFACATE_FILEPATH, std::string(msg->response_body->data, msg->response_body->length));
            if (!this->generate_certificate_file())
            {
                g_object_unref(msg);
                return false;
            }
            this->enable_repo_groups(true);
        }
        else
        {
            KLOG_WARNING("Failed to request certifacate, status code: %d.", status);
            g_object_unref(msg);
            return false;
        }
    }
    catch (const Glib::Error& e)
    {
        KLOG_WARNING("%s", e.what().c_str());
        g_object_unref(msg);
        return false;
    }

    g_object_unref(msg);
    return true;
#endif
    return false;
}

void YumRepoControl::on_dnf_vars_changed(const Glib::RefPtr<Gio::File>& file,
                                         const Glib::RefPtr<Gio::File>& other_file,
                                         Gio::FileMonitorEvent event_type)
{
    if (event_type == Gio::FILE_MONITOR_EVENT_CHANGED || Gio::FILE_MONITOR_EVENT_CREATED)
    {
        this->load_dnf_vars();
        // osversion文件变化可能是用户需要升级大版本，因此需要重新获取新版本的证书文件
        this->download_certificate_file();
    }
}

void YumRepoControl::on_certificate_file_changed(const Glib::RefPtr<Gio::File>& file,
                                                 const Glib::RefPtr<Gio::File>& other_file,
                                                 Gio::FileMonitorEvent event_type)
{
    if (event_type == Gio::FILE_MONITOR_EVENT_DELETED)
    {
        this->certificate_file_deleted_.emit();
        KLOG_WARNING("certificate file deleted!");
    }
}

bool YumRepoControl::generate_certificate_file()
{
    try
    {
        auto command_line = fmt::format("/usr/bin/tar -zxvf {0} -C {1}", CERTIFACATE_FILEPATH, KL_INSTALL_DATADIR);
        std::string standard_error;
        std::string standard_output;
        Glib::spawn_command_line_sync(command_line, &standard_output, &standard_error);
        if (standard_error.length() > 0)
        {
            KLOG_WARNING("%s", standard_error.c_str());
            return false;
        }
        else
        {
            Glib::Dir certificate_file_dir(KL_INSTALL_DATADIR "/" REPO_NAME);
            std::vector<std::string> certificate_files;
            while (true)
            {
                auto certificate_file_name = certificate_file_dir.read_name();
                if (certificate_file_name == "")
                {
                    break;
                }

                auto file_name_splits = Base::StrUtils::split_with_char(certificate_file_name, '.');
                if (file_name_splits.size() == 2 && file_name_splits[1] == std::string("key"))
                {
                    command_line = fmt::format("/usr/bin/mv {0} {1}",
                                               (std::string(KL_INSTALL_DATADIR) + "/" + REPO_NAME + "/" + certificate_file_name).c_str(),
                                               std::string(KL_INSTALL_DATADIR) + "/" + this->osversion_ + ".key");
                    KLOG_DEBUG("%s", command_line.c_str());
                    Glib::spawn_command_line_sync(command_line);
                }
                else if (file_name_splits.size() == 2 && file_name_splits[1] == std::string("crt"))
                {
                    command_line = fmt::format("/usr/bin/mv {0} {1}",
                                               (std::string(KL_INSTALL_DATADIR) + "/" + REPO_NAME + "/" + certificate_file_name).c_str(),
                                               std::string(KL_INSTALL_DATADIR) + "/" + this->osversion_ + ".crt");
                    KLOG_DEBUG("%s", command_line.c_str());
                    Glib::spawn_command_line_sync(command_line);
                }
            }
            command_line = fmt::format("/usr/bin/rm -rf {0}",
                                       (std::string(KL_INSTALL_DATADIR) + "/" + REPO_NAME).c_str());
            Glib::spawn_command_line_sync(command_line);
        }
    }
    catch (const Glib::Error& e)
    {
        KLOG_WARNING("%s", e.what().c_str());
        return false;
    }
    return true;
}

void YumRepoControl::enable_repo_groups(bool enable)
{
#ifdef ENABLE_SERVER
    if (!Glib::file_test(this->repo_file_name_, Glib::FileTest::FILE_TEST_EXISTS))
    {
        KLOG_WARNING("repo file can't exist");
        return;
    }

    try
    {
        Glib::KeyFile repo_config;
        repo_config.load_from_file(this->repo_file_name_);
        for (auto iter = this->repo_group_names_.begin(); iter != this->repo_group_names_.end(); iter++)
        {
            if (repo_config.has_group(*iter))
            {
                repo_config.set_integer(*iter, REPO_KEY_ENABLE, enable ? REPO_ENABLE_VALUE : REPO_DISABLE_VALUE);
                repo_config.set_integer(*iter, REPO_KEY_SSLVERIFY, 0);
                repo_config.set_string(*iter, REPO_KEY_SSLCLIENTCERT, KL_INSTALL_DATADIR REPO_SSLCLIENTCERT_FILE_NAME);
                repo_config.set_string(*iter, REPO_KEY_SSLCLIENTKEY, KL_INSTALL_DATADIR REPO_SSLCLIENTKEY_FILE_NAME);
            }
            else
            {
                KLOG_WARNING("group %s can't exist", iter->c_str());
                return;
            }
        }

        auto repo_data = repo_config.to_data();
        Base::FileUtils::write_bdata(this->repo_file_name_, repo_data, O_WRONLY | O_CREAT | O_TRUNC | O_SYNC, LICENSE_FILE_PERMISSION);
    }
    catch (const Glib::Error& e)
    {
        KLOG_WARNING("%s.", e.what().c_str());
    }
#endif
}

void YumRepoControl::enable_repo_groups(std::string repo_name, bool enable)
{
#ifdef ENABLE_SERVER
    if (!Glib::file_test(this->repo_file_name_, Glib::FileTest::FILE_TEST_EXISTS))
    {
        KLOG_WARNING("repo file can't exist");
        return;
    }

    try
    {
        Glib::KeyFile repo_config;
        repo_config.load_from_file(this->repo_file_name_);

        if (repo_config.has_group(repo_name))
        {
            repo_config.set_integer(repo_name, REPO_KEY_ENABLE, enable ? REPO_ENABLE_VALUE : REPO_DISABLE_VALUE);
        }
        else
        {
            KLOG_WARNING("group %s can't exist", repo_name.c_str());
            return;
        }

        auto repo_data = repo_config.to_data();
        Base::FileUtils::write_bdata(this->repo_file_name_, repo_data, O_WRONLY | O_CREAT | O_TRUNC | O_SYNC, LICENSE_FILE_PERMISSION);
    }
    catch (const Glib::Error& e)
    {
        KLOG_WARNING("%s.", e.what().c_str());
    }
#endif
}

}  // namespace Daemon