/**
 * @file          /kylin-license/src/register/license-register.cpp
 * @brief
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */

#include "src/register/license-register.h"

#include <glib/gi18n.h>

namespace Register
{

#define DBUS_DAEMON_DBUS_NAME "org.freedesktop.DBus"
#define DBUS_DAEMON_OBJECT_PATH "/org/freedesktop/DBus"
#define DBUS_DAEMON_FUNCTION_NAME_HAS_OWNER "NameHasOwner"

LicenseRegister::LicenseRegister(::DBus::Connection dbus_connection) : option_group_("config", "config options"),
                                                                       dbus_connection_(dbus_connection)
{
}

LicenseRegister::~LicenseRegister()
{
}

bool LicenseRegister::init()
{
    this->option_group_.add_entry(Base::MiscUtils::create_option_entry('r', "register",
                                                                       N_("Activate the machine. If the activation code is specified, then activate machine using activation code,"
                                                                          " and if the server host is specified, then activate machine online. Otherwise, activate using USBKey device.")),
                                  this->options_.registration);
    this->option_group_.add_entry(Base::MiscUtils::create_option_entry('l', "license-code",
                                                                       N_("Specify the activation code, the argument is deprecated, please use -a.")),
                                  this->options_.license_code);
    this->option_group_.add_entry(Base::MiscUtils::create_option_entry('a', "activation-code",
                                                                       N_("Specify the activation code.")),
                                  this->options_.activation_code);
    this->option_group_.add_entry(Base::MiscUtils::create_option_entry('d', "server-host",
                                                                       N_("Specify the activation server host.")),
                                  this->options_.server_host);

    this->option_group_.add_entry(Base::MiscUtils::create_option_entry("service-uid",
                                                                       N_("Specify the service uid for activation server.")),
                                  this->options_.service_uid);

    this->option_group_.add_entry(Base::MiscUtils::create_option_entry('s', "show",
                                                                       N_("Show activation info.")),
                                  this->options_.show_license);
    this->option_group_.add_entry(Base::MiscUtils::create_option_entry('v', "version",
                                                                       N_("Output version information and exit.")),
                                  this->options_.show_version);
    this->option_group_.add_entry(Base::MiscUtils::create_option_entry('g', "get",
                                                                       N_("Get the specified variable."), "registration-status|status|expired-status|machinecode|customercode|authcode|expiredate"),
                                  this->options_.variable_name);
    this->option_group_.add_entry(Base::MiscUtils::create_option_entry('o', "activation-object-name",
                                                                       N_("Name of activation object. If not specified, KylinSecOS is used.")),
                                  this->options_.license_object_name);

    this->option_group_.add_entry(Base::MiscUtils::create_option_entry('j', "json",
                                                                       N_("Use JSON output format.")),
                                  this->options_.show_by_json);
    this->option_group_.add_entry(Base::MiscUtils::create_option_entry("deactivate",
                                                                       N_("Deactivate.")),
                                  this->options_.deactive);
    this->option_group_.add_entry(Base::MiscUtils::create_option_entry("wait-service",
                                                                       N_("Wait for ksl-daemon service start.")),
                                  this->options_.wait_service);

    this->option_group_.set_translation_domain(PROJECT_NAME);
    this->option_context_.set_main_group(this->option_group_);

    try
    {
        this->license_manager_proxy_ = DBus::LicenseManagerProxyWrapper::create(this->dbus_connection_, LICENSE_MANAGER_OBJECT_PATH, LICENSE_MANAGER_DBUS_NAME);
        // TODO:
        // this->license_manager_proxy_->dbusProxy()->set_default_timeout(LICENSE_DBUS_TIMEOUT_MSEC);
    }
    catch (const std::exception& e)
    {
        KLOG_WARNING("%s", e.what());
        fmt::print(stderr, "{0}", e.what());
        return false;
    }

    return true;
}

bool LicenseRegister::run(int& argc, char**& argv)
{
    try
    {
        this->option_context_.parse(argc, argv);
    }
    catch (const Glib::Exception& e)
    {
        KLOG_WARNING("%s", e.what().c_str());
        fmt::print(stderr, _("Failed to parse command arguments\n"));
        return false;
    }

    try
    {
        if (false == this->options_.wait_service)
        {
            /*Get "service com.kylinsec.Kiran.LicenseHelper" status* BEGIN: 防止开机时ky-verify服务未启动导致的桌面无法启动问题*/
            auto sys_bus = ::DBus::Connection::SystemBus();
            auto call = ::DBus::CallMessage(DBUS_DAEMON_DBUS_NAME,
                                            DBUS_DAEMON_OBJECT_PATH,
                                            DBUS_DAEMON_DBUS_NAME,
                                            DBUS_DAEMON_FUNCTION_NAME_HAS_OWNER);
            ::DBus::MessageIter wi = call.writer();
            wi << std::string(LICENSE_DBUS_NAME);

            ::DBus::Message ret = sys_bus.send_blocking(call, 1000);
            ::DBus::MessageIter recive = ret.reader();
            bool arg_out;
            recive >> arg_out;
            if (!arg_out)
            {
                KLOG_ERROR("DBUS \'service com.kylinsec.Kiran.LicenseHelper\' can't exitst!");
                return false;
            }
            /*Get "service com.kylinsec.Kiran.LicenseHelper" status* END:*/
        }
    }
    catch (const std::exception& e)
    {
        KLOG_WARNING("%s", e.what());
        fmt::print(stderr, "{0}", e.what());
        return false;
    }

    if (this->options_.show_version)
    {
        fmt::print("Kylin license 2.6\n");
        return true;
    }

    // 兼容旧的版本
    if (this->options_.license_object_name.empty())
    {
        this->options_.license_object_name = LICENSE_OBJECT_OS_NAME;
    }

    try
    {
        auto object_path = this->license_manager_proxy_->GetLicenseObject(this->options_.license_object_name);
        this->license_object_proxy_ = DBus::LicenseObjectProxyWrapper::create(this->dbus_connection_, object_path, LICENSE_MANAGER_DBUS_NAME);

        if (this->options_.deactive)
        {
            if (this->deactive_license())
            {
                fmt::print(_("Suceess deactive license\n"));
            }
            else
            {
                fmt::print(_("Failed deactive license\n"));
            }
        }

        if (!this->options_.variable_name.empty())
        {
            this->process_get_variable(this->options_.variable_name);
            return true;
        }

        if (this->options_.registration)
        {
            // -a和-l选项相同，优先使用-a选项，如果-a未设置但-l设置了才使用-l选项的参数
            if (!this->options_.license_code.empty() && this->options_.activation_code.empty())
            {
                this->options_.activation_code = this->options_.license_code;
            }

            if (!this->process_registration(this->options_.activation_code,
                                            this->options_.server_host,
                                            this->options_.service_uid))
            {
                return false;
            }
        }

        if (this->options_.show_license)
        {
            this->process_show();
        }
    }
    catch (const std::exception& e)
    {
        fmt::print(stderr, "{0}", e.what());
        return false;
    }

    return true;
}

/**
 * @brief kyreg -g 提供给云桌面用的接口
 * @param variable_name
 */
void LicenseRegister::process_get_variable(const std::string& variable_name)
{
    KLOG_PROFILE("variable name: %s", variable_name.c_str());

    Json::Value license_value;
    std::string error;

    if (!this->get_license(license_value, error))
    {
        fmt::print(stderr, "{0}", error);
        return;
    }

    auto activation_status = license_value["activation_status"].asUInt();
    auto service_status = license_value["service_status"].asUInt();
    auto machine_code = license_value["machine_code"].asString();
    auto customer_info = license_value["customer_info"].asString();
    auto activation_code = license_value["activation_code"].asString();
    auto expired_time = license_value["expired_time"].asUInt64();
    auto expired_gdate = Base::TimeUtils::timestamp2gdate(expired_time);
    auto expired_date_str = Base::TimeUtils::gdate2str(&expired_gdate);

    if (variable_name == "registration-status")
    {
        fmt::print("{0}\n", activation_status == LicenseActivationStatus::LAS_UNACTIVATED ? "unregistered" : "registered");
    }
    else if (variable_name == "status")
    {
        fmt::print("{0}\n", activation_status == LicenseActivationStatus::LAS_UNACTIVATED ? "inactive" : "activate");
    }
    else if (variable_name == "expired-status")
    {
        if (service_status == LicenseServiceStatus::LICENSE_SERVICE_STATUS_UNEXPIRED)
        {
            fmt::print("unexpired\n");
        }
        else
        {
            fmt::print("expired\n");
        }
    }
    else if (variable_name == "machinecode")
    {
        fmt::print("{0}\n", machine_code);
    }
    else if (variable_name == "customercode")
    {
        fmt::print("{0}\n", customer_info);
    }
    else if (variable_name == "authcode")
    {
        fmt::print("{0}\n", activation_code);
    }
    else if (variable_name == "expiredate")
    {
        fmt::print("{0}\n", (g_date_year(&expired_gdate) >= 2250) ? std::string("Forever") : expired_date_str);
    }
}

bool LicenseRegister::process_registration(const std::string& activation_code,
                                           const std::string& auth_server,
                                           const std::string& service_uid)
{
    KLOG_PROFILE("License code: %s auth server: %s",
                 activation_code.c_str(),
                 auth_server.c_str());

    std::string error;

    RETURN_VAL_IF_FALSE(this->license_object_proxy_, false);

    try
    {
        // 使用激活码激活
        if (activation_code.length() > 0)
        {
            this->license_object_proxy_->ActivateByActivationCode(activation_code);
        }
        // 在线激活
        else if (auth_server.length() > 0)
        {
            if (service_uid.length() == 0)
            {
                fmt::print(stderr, _("Service uid is required."));
                return false;
            }
            else
            {
                this->license_object_proxy_->ActivateOnLine(auth_server, service_uid);
            }
        }
        // UKEY激活
        else
        {
            auto result = this->license_object_proxy_->ActivateByUsbKey();

            Json::Value ukey_value;

            if (!this->json_str2value(result, ukey_value, error))
            {
                fmt::print(stderr, "{0}", error);
                return false;
            }

            fmt::print(_("Sum points:            {0}\n"), ukey_value[LICENSE_JK_SUM_POINTS].asInt());
            fmt::print(_("Remaining points:      {0}\n"), ukey_value[LICENSE_JK_REMAINING_POINTS].asInt());
        }
    }
    catch (const Glib::Error& e)
    {
        fmt::print(stderr, "{0}", e.what().c_str());
        return false;
    }
    return true;
}

void LicenseRegister::process_show()
{
    Json::Value license_value;
    std::string error;

    if (!this->get_license(license_value, error))
    {
        fmt::print(stderr, "{0}", error);
        return;
    }

    // 这里主要是提供给上层应用使用，方便上层应用通过命令行获取激活信息
    if (this->options_.show_by_json)
    {
        auto json_str = Base::StrUtils::json2str(license_value);
        fmt::print("{0}", json_str.c_str());
        return;
    }

    auto activation_status = license_value[LICENSE_JK_ACTIVATION_STATUS].asUInt();
    auto service_status = license_value[LICENSE_JK_SERVICE_STATUS].asUInt();
    auto activation_type = license_value[LICENSE_JK_ACTIVATION_TYPE].asUInt();
    auto activation_time = license_value[LICENSE_JK_ACTIVATION_TIME].asUInt64();
    auto expired_time = license_value[LICENSE_JK_EXPIRED_TIME].asUInt64();
    auto first_boot_time = license_value[LICENSE_JK_FIRST_BOOT_TIME].asUInt64();
    auto machine_code = license_value[LICENSE_JK_MACHINE_CODE].asString();
    auto activation_code = license_value[LICENSE_JK_ACTIVATION_CODE].asString();
    auto customer_info = license_value[LICENSE_JK_CUSTOMER_INFO].asString();
    auto update_mode = license_value[LICENSE_JK_UPDATE_MODE].asUInt();
    auto service_mode = license_value[LICENSE_JK_SERVICE_MODE].asUInt();
    auto exception_status = license_value[LICENSE_JK_EXCEPTION_STATUS].asUInt();

    auto expired_gdate = Base::TimeUtils::timestamp2gdate(expired_time);
    auto expired_date_str = Base::TimeUtils::gdate2str(&expired_gdate);
    if (activation_status == LicenseActivationStatus::LAS_ACTIVATED)
    {
        auto register_type_str = this->get_activation_type_str(activation_type);

        auto register_date_str = Base::TimeUtils::timestamp2str(activation_time);
        fmt::print(_("Activation status:    Activated.\n"));
        if (exception_status != LicenseExceptionStatus::LICENSE_EXCEPTION_STATUS_NONE)
        {
            fmt::print(_("Exception infomation: {0}.\n"), this->get_exception_string(exception_status));
        }
        fmt::print(_("Expiry date:          {0}.\n"), (g_date_year(&expired_gdate) >= 2250) ? _("Forever") : expired_date_str);
        fmt::print(_("Activation type:      {0}.\n"), register_type_str);
        fmt::print(_("Activation code:      {0}.\n"), activation_code);
        if (update_mode != LicenseUpdateMode::LICENSE_UPDATE_MODE_NONE)
        {
            fmt::print(_("Update mode:          {0}.\n"), update_mode == LicenseUpdateMode::LICENSE_UPDATE_MODE_PERMANENT ? _("Permanent") : _("Subscribe"));
        }

        if (service_mode != LicenseServiceMode::LICENSE_SERVICE_MODE_NONE)
        {
            fmt::print(_("Service mode:         {0}.\n"), service_mode == LicenseServiceMode::LICENSE_SERVICE_MODE_7X24 ? _("7x24 hours") : _("5x8 hours"));
        }
        fmt::print(_("Machine code:         {0}.\n"), machine_code);
        fmt::print(_("Activation time:      {0}.\n"), register_date_str);
    }
    else
    {
        // LICENSE_EXCEPTION_STATUS_INVALID异常情况只存在于未激活的机器
        if (exception_status == LicenseExceptionStatus::LICENSE_EXCEPTION_STATUS_INVALID)
        {
            fmt::print(_("Activation status:    The current time is illegal. (less than installation time: {0}).\n"),
                       Base::TimeUtils::timestamp2str(first_boot_time));
        }
        else
        {
            fmt::print(_("Activation status:    Not activated.\n"));
            if (exception_status != LicenseExceptionStatus::LICENSE_EXCEPTION_STATUS_NONE)
            {
                fmt::print(_("Exception infomation: {0}.\n"), this->get_exception_string(exception_status));
            }
            // 只有系统激活才显示试用期
            if (this->options_.license_object_name == LICENSE_OBJECT_OS_NAME)
            {
                fmt::print(_("Trial expiration:  {0}.\n"), expired_date_str);
            }
        }

        fmt::print(_("Machine code:         {0}.\n"), machine_code);
    }

    // 只有系统激活才显示安装时间
    if (this->options_.license_object_name == LICENSE_OBJECT_OS_NAME)
    {
        auto start_date_str = Base::TimeUtils::timestamp2str(first_boot_time);
        fmt::print(_("Installation time:    {0}.\n"), start_date_str);
    }
}

bool LicenseRegister::get_license(Json::Value& value, std::string& error)
{
    KLOG_PROFILE("");

    if (!this->license_object_proxy_)
    {
        error = _("The dbus proxy of license object is uninitialized");
        return false;
    }

    try
    {
        auto license_json = this->license_object_proxy_->GetLicense();
        return this->json_str2value(license_json, value, error);
    }
    catch (const Glib::Error& e)
    {
        error = e.what();
        return false;
    }
}

bool LicenseRegister::deactive_license()
{
    if (!this->license_object_proxy_)
    {
        KLOG_ERROR("The dbus proxy of license object is uninitialized");
        return false;
    }

    try
    {
        return this->license_object_proxy_->DeactiveLicense();
    }
    catch (const Glib::Error& e)
    {
        KLOG_ERROR(e.what().c_str());
        return false;
    }
}

bool LicenseRegister::json_str2value(const std::string& str, Json::Value& value, std::string& error)
{
    Json::CharReaderBuilder reader_builder;
    std::unique_ptr<Json::CharReader> reader(reader_builder.newCharReader());
    return reader->parse(str.c_str(), str.c_str() + str.length(), &value, &error);
}

std::string LicenseRegister::get_activation_type_str(uint32_t activation_type)
{
    switch (activation_type)
    {
    case LicenseActivationType::LICENSE_ACTIVATION_TYPE_MANNUAL:
        return std::string(_("Activation code"));

    case LicenseActivationType::LICENSE_ACTIVATION_TYPE_ONLINE:
        return std::string(_("Network"));

    case LicenseActivationType::LICENSE_ACTIVATION_TYPE_OFFLINE:
        return std::string(_("USBKey"));

    default:
        return std::string(_("Unknown"));
    }
    return std::string(_("Unknown"));
}

std::string LicenseRegister::get_exception_string(uint32_t exception_status)
{
    std::string exception_string;
    switch (exception_status)
    {
    case LicenseExceptionStatus::LICENSE_EXCEPTION_STATUS_CS_TIME_DIFF_TOO_MUCH:
        exception_string = _("The difference between the current time and the server time is too large, please check the system time");
        break;
    case LicenseExceptionStatus::LICENSE_EXCEPTION_STATUS_SERVER_NOT_CONNECTED:
        exception_string = _("Communication failure with the server, please check the network");
        break;
    default:
        break;
    }
    return exception_string;
}

}  // namespace Register
