/**
 * @file          /ks-license/src/server/license-database.cpp
 * @brief
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */

#include "src/server/license-database.h"
#include <iostream>
#include "lib/base/base.h"

namespace Server
{
#define DB_FILE_PATH KL_INSTALL_DATADIR "/points.db"

#define CREATE_ACTIVATION_TABLE_IF_NO_EXIST                                \
    "CREATE TABLE IF NOT EXISTS license_activation_info(\n"                \
    "uid                      INTEGER        PRIMARY KEY AUTOINCREMENT,\n" \
    "service_uid              VARCHAR(32)    NOT NULL,\n"                  \
    "activation_object_name   VARCHAR(32)    NOT NULL,\n"                  \
    "machine_code             VARCHAR(32)    NOT NULL,\n"                  \
    "activation_code          VARCHAR(64)    NOT NULL,\n"                  \
    "expiry_date              DATE           NOT NULL,\n"                  \
    "activation_time          DATETIME       NOT NULL,\n"                  \
    "keep_alive_timestamp     INTEGER        DEFAULT 0,\n"                 \
    "client_ip                VARCHAR(32)    NOT NULL,\n"                  \
    "UNIQUE(machine_code, service_uid));\n"

#define CREATE_POINTS_TABLE_IF_NO_EXIST                                   \
    "CREATE TABLE IF NOT EXISTS license_points_info(\n"                   \
    "uid                               VARCHAR(32)        PRIMARY KEY,\n" \
    "contract_no                       VARCHAR(32)        NOT NULL,\n"    \
    "subject_name                      VARCHAR(256)       NOT NULL,\n"    \
    "customer_info                     VARCHAR(32)        NOT NULL,\n"    \
    "activation_object_name            VARCHAR(32)        NOT NULL,\n"    \
    "expiry_date                       DATE               NOT NULL,\n"    \
    "used_on_virtual_machine           INT                DEFAULT 0,\n"   \
    "keep_alive_time                   INT                DEFAULT 0,\n"   \
    "enable_recycling                  BOOLEAN            DEFAULT 0,\n"   \
    "points_activation_code            VARCHAR(64),\n"                    \
    "install_type                      INT                NOT NULL,\n"    \
    "update_mode                       INT                DEFAULT 0,\n"   \
    "service_mode                      INT                DEFAULT 0,\n"   \
    "sum_points                        INT                NOT NULL,\n"    \
    "remaining_points                  INT                NOT NULL,\n"    \
    "activation_object_note            VARCHAR(32)        DEFAULT '',\n"  \
    "warranty_days                     INT                DEFAULT 0,\n"   \
    "compute_expiry_day_mode           INT                DEFAULT 0,\n"   \
    "warranty_years                    INT                DEFAULT 0,\n"   \
    "warranty_months                   INT                DEFAULT 0,\n"   \
    "customization_customer            VARCHAR(64)        DEFAULT '');\n"

#define SELECT_ACTIVATION_INFO_BY_SERVICE_UID "SELECT uid, service_uid, activation_object_name, machine_code, "                 \
                                              "activation_code, expiry_date, activation_time, keep_alive_timestamp, client_ip " \
                                              "from license_activation_info where service_uid = '{0}';"

#define REPLACE_ACTIVATION_RECORD                                                                                                         \
    "REPLACE INTO license_activation_info("                                                                                               \
    "service_uid, activation_object_name, machine_code, activation_code, expiry_date, activation_time, keep_alive_timestamp, client_ip) " \
    "VALUES('{0}', '{1}', '{2}', '{3}', DATE('{4}'), DATETIME('{5}'), {6}, '{7}');"

#define DELETE_ACTIVATION_RECORD "DELETE FROM license_activation_info WHERE machine_code = '{0}';"

#define DELETE_POINTS_RECORD "DELETE FROM license_points_info WHERE uid = '{0}';"

#define SELECT_ALL_POINTS_INFO                                                                                                                   \
    "SELECT uid, contract_no, subject_name, customer_info, activation_object_name, expiry_date, used_on_virtual_machine, keep_alive_time,"       \
    " enable_recycling, points_activation_code, install_type, update_mode, service_mode, sum_points, remaining_points, activation_object_note, " \
    "warranty_days, compute_expiry_day_mode, warranty_years, warranty_months, customization_customer from license_points_info;"

#define UPDATE_POINTS_RECORD "UPDATE license_points_info SET remaining_points = {0} WHERE uid = '{1}';"

#define UPDATE_POINTS_EXPIRY_DAY "UPDATE license_points_info SET expiry_date = '{0}' WHERE uid = '{1}';"

#define UPDATE_POINTS_PERIOD_DAY "UPDATE license_points_info SET warranty_days = '{0}' WHERE uid = '{1}';"

#define UPDATE_POINTS_PERIOD_MONTH "UPDATE license_points_info SET warranty_months = '{0}' WHERE uid = '{1}';"

#define UPDATE_POINTS_PERIOD_YEAR "UPDATE license_points_info SET warranty_years = '{0}' WHERE uid = '{1}';"

#define INSERT_POINTS_RECORD "INSERT INTO license_points_info("                                                                                               \
                             "uid, contract_no, subject_name, customer_info, activation_object_name, expiry_date, used_on_virtual_machine, keep_alive_time, " \
                             "enable_recycling, points_activation_code, install_type, update_mode, service_mode, sum_points, remaining_points, "              \
                             "activation_object_note, warranty_days, compute_expiry_day_mode, warranty_years, warranty_months, customization_customer)"       \
                             "VALUES('{0}', '{1}', '{2}','{3}','{4}','{5}', {6}, {7}, '{8}', '{9}', {10}, {11}, {12}, {13}, {14},'{15}', {16}, {17}, {18}, {19}, '{20}')"

#define ADD_COLUMN(tabel_name, column_name, value_type, default_value) "\nALTER TABLE " + table_name + " ADD COLUMN " + column_name + " " + value_type + " DEFAULT " + default_value + ";"

#define IS_COLUMN_EXITS(tabel_name, column_name) "SELECT * FROM sqlite_master WHERE NAME='" + table_name + "' AND SQL LIKE '%%" + column_name + "%%';"

LicenseDatabase::LicenseDatabase() : file_path_(DB_FILE_PATH),
                                     db_handle_(NULL)
{
}

LicenseDatabase::~LicenseDatabase()
{
}

std::shared_ptr<LicenseDatabase> LicenseDatabase::instance_ = std::shared_ptr<LicenseDatabase>();
std::shared_ptr<LicenseDatabase> LicenseDatabase::get_default()
{
    if (!instance_)
    {
        instance_ = std::make_shared<LicenseDatabase>();
        instance_->init();
    }
    return instance_;
}

void LicenseDatabase::init()
{
    KLOG_PROFILE("");

    RETURN_IF_TRUE(this->file_path_.length() == 0);

    if (!Glib::file_test(this->file_path_, Glib::FILE_TEST_EXISTS))
    {
        auto dirname = Glib::path_get_dirname(this->file_path_);
        g_mkdir_with_parents(dirname.c_str(), NEW_FILE_PERMISSION);
    }

    // 打开数据库
    auto ret = sqlite3_open(this->file_path_.c_str(), &this->db_handle_);
    if (ret != SQLITE_OK)
    {
        KLOG_WARNING("%s, db_path: %s", sqlite3_errmsg(this->db_handle_), this->file_path_.c_str());
        return;
    }

    if (sqlite3_key(this->db_handle_, DB_KEY, strlen(DB_KEY)) != SQLITE_OK)
    {
        KLOG_WARNING("Failed to decrypt db.");
        return;
    }

    std::string error;

    this->process_column_compatible("license_points_info", "activation_object_note", "VARCHAR(32)", "\'\'");
    this->process_column_compatible("license_points_info", "warranty_days", "INT", "0");
    this->process_column_compatible("license_points_info", "compute_expiry_day_mode", "INT", "0");
    this->process_column_compatible("license_points_info", "warranty_years", "INT", "0");
    this->process_column_compatible("license_points_info", "warranty_months", "INT", "0");
    this->process_column_compatible("license_points_info", "customization_customer", "VARCHAR(64)", "\'\'");

    // 创建授权表
    if (!this->exec_sql(CREATE_ACTIVATION_TABLE_IF_NO_EXIST, error))
    {
        KLOG_WARNING("%s", error.c_str());
    }

    // 创建点数记录表
    if (!this->exec_sql(CREATE_POINTS_TABLE_IF_NO_EXIST, error))
    {
        KLOG_WARNING("%s", error.c_str());
    }
}

PointsRecordVec LicenseDatabase::get_points_records()
{
    PointsRecordVec points_records;
    std::string error;

    if (!this->exec_sql(SELECT_ALL_POINTS_INFO, error, &LicenseDatabase::parse_points_record, &points_records))
    {
        KLOG_WARNING("%s", error.c_str());
        return PointsRecordVec();
    }
    return points_records;
}

bool LicenseDatabase::update_points_record(std::shared_ptr<PointsRecord> points_record)
{
    RETURN_VAL_IF_FALSE(points_record, false);

    auto sql = fmt::format(UPDATE_POINTS_RECORD, points_record->remaining_points, points_record->uid);

    std::string error;
    if (!this->exec_sql(sql, error))
    {
        KLOG_WARNING("%s", error.c_str());
        return false;
    }
    return true;
}

bool LicenseDatabase::update_points_expiry_day(const std::string &uid, const std::string &expiry_day)
{
    auto sql = fmt::format(UPDATE_POINTS_EXPIRY_DAY, expiry_day, uid);

    std::string error;
    if (!this->exec_sql(sql, error))
    {
        KLOG_WARNING("%s", error.c_str());
        return false;
    }
    return true;
}

bool LicenseDatabase::update_points_by_period(const std::string &uid, int32_t warranty_days, int32_t warranty_months, int32_t warranty_years)
{
    std::string sql;
    std::string error;

    sql = fmt::format(UPDATE_POINTS_PERIOD_DAY, warranty_days, uid);
    if (!this->exec_sql(sql, error))
    {
        KLOG_WARNING("%s", error.c_str());
        return false;
    }

    sql = fmt::format(UPDATE_POINTS_PERIOD_MONTH, warranty_months, uid);
    if (!this->exec_sql(sql, error))
    {
        KLOG_WARNING("%s", error.c_str());
        return false;
    }

    sql = fmt::format(UPDATE_POINTS_PERIOD_YEAR, warranty_years, uid);
    if (!this->exec_sql(sql, error))
    {
        KLOG_WARNING("%s", error.c_str());
        return false;
    }

    return true;
}

ActivationRecordVec LicenseDatabase::get_activation_records(const std::string &service_uid)
{
    ActivationRecordVec activation_records;
    std::string error;

    auto sql = fmt::format(SELECT_ACTIVATION_INFO_BY_SERVICE_UID, service_uid);
    if (!this->exec_sql(sql, error, &LicenseDatabase::parse_activation_record, &activation_records))
    {
        KLOG_WARNING("%s", error.c_str());
        return ActivationRecordVec();
    }
    return activation_records;
}

bool LicenseDatabase::update_activation_record(std::shared_ptr<ActivationRecord> activation_record)
{
    auto sql = fmt::format(REPLACE_ACTIVATION_RECORD,
                           activation_record->service_uid,
                           activation_record->activation_object_name,
                           activation_record->machine_code,
                           activation_record->activation_code,
                           activation_record->expiry_date,
                           activation_record->activation_time,
                           activation_record->keep_alive_timestamp,
                           activation_record->client_ip);

    std::string error;
    if (!this->exec_sql(sql, error))
    {
        KLOG_WARNING("%s", error.c_str());
        return false;
    }
    return true;
}

bool LicenseDatabase::delete_activation_record(const std::string &machine_code)
{
    auto sql = fmt::format(DELETE_ACTIVATION_RECORD, machine_code);

    std::string error;
    if (!this->exec_sql(sql, error))
    {
        KLOG_WARNING("%s", error.c_str());
        return false;
    }
    return true;
}

int LicenseDatabase::parse_activation_record(void *user_data, int column_num, char **values, char **names)
{
    ActivationRecordVec *activation_records = (ActivationRecordVec *)(user_data);
    RETURN_VAL_IF_TRUE(activation_records == NULL, -1);
    RETURN_VAL_IF_TRUE(column_num < 9, -1);
    auto activation_record = std::make_shared<ActivationRecord>();
    activation_record->service_uid = POINTER_TO_STRING(values[1]);
    activation_record->activation_object_name = POINTER_TO_STRING(values[2]);
    activation_record->machine_code = POINTER_TO_STRING(values[3]);
    activation_record->activation_code = POINTER_TO_STRING(values[4]);
    activation_record->expiry_date = POINTER_TO_STRING(values[5]);
    activation_record->activation_time = POINTER_TO_STRING(values[6]);
    activation_record->keep_alive_timestamp = uint32_t(strtoul(values[7], NULL, 0));
    activation_record->client_ip = POINTER_TO_STRING(values[8]);
    activation_records->push_back(activation_record);

    return SQLITE_OK;
}

int LicenseDatabase::parse_points_record(void *user_data, int column_num, char **values, char **names)
{
    PointsRecordVec *points_records = (PointsRecordVec *)(user_data);
    RETURN_VAL_IF_TRUE(points_records == NULL, -1);
    RETURN_VAL_IF_TRUE(column_num < 16, -1);
    auto points_record = std::make_shared<PointsRecord>();
    points_record->uid = POINTER_TO_STRING(values[0]);
    points_record->contract_no = POINTER_TO_STRING(values[1]);
    points_record->subject_name = POINTER_TO_STRING(values[2]);
    points_record->customer_info = POINTER_TO_STRING(values[3]);
    points_record->activation_object_name = POINTER_TO_STRING(values[4]);
    points_record->expiry_date = POINTER_TO_STRING(values[5]);
    points_record->used_on_virtual_machine = strtol(values[6], NULL, 0);
    points_record->keep_alive_time = strtol(values[7], NULL, 0);
    points_record->enable_recycling = (strcmp(values[8], "true") == 0);
    points_record->points_activation_code = POINTER_TO_STRING(values[9]);
    points_record->install_type = strtol(values[10], NULL, 0);
    points_record->update_mode = strtol(values[11], NULL, 0);
    points_record->service_mode = strtol(values[12], NULL, 0);
    points_record->sum_points = strtol(values[13], NULL, 0);
    points_record->remaining_points = strtol(values[14], NULL, 0);
    points_record->activation_object_note = POINTER_TO_STRING(values[15]);
    points_record->warranty_days = strtol(values[16], NULL, 0);
    points_record->compute_expiry_day_mode = strtol(values[17], NULL, 0);
    points_record->warranty_years = strtol(values[18], NULL, 0);
    points_record->warranty_months = strtol(values[19], NULL, 0);
    points_record->customization_customer = POINTER_TO_STRING(values[20]);
    points_records->push_back(points_record);

    return SQLITE_OK;
}

int LicenseDatabase::check_sql_return(void *user_data, int column_num, char **values, char **names)
{
    bool *is_return_value = (bool *)user_data;
    column_num < 1 ? *is_return_value = false : *is_return_value = true;

    return SQLITE_OK;
}

bool LicenseDatabase::exec_sql(std::string sql, std::string &error, sqlite3_callback callback, void *user_data)
{
    RETURN_VAL_IF_TRUE(this->db_handle_ == NULL, true);

    std::unique_lock<std::mutex> lock(this->mutex_);

    char *sqlite3_error = NULL;
    auto ret = sqlite3_exec(this->db_handle_, sql.c_str(), callback, user_data, &sqlite3_error);
    if (ret != SQLITE_OK)
    {
        error = fmt::format("Failed to exec sql {0}. error: {1}.", sql, POINTER_TO_STRING(sqlite3_error).c_str());
        sqlite3_free(sqlite3_error);
        return false;
    }
    KLOG_DEBUG("Success to exec sql: %s.", sql.c_str());
    return true;
}

PointsRecordVec LicenseDatabase::merge_points_file(const std::string &file_name, LicenseErrorCode &error_code)
{
    auto new_points_records = this->get_points_records_by_file(file_name);
    error_code = LicenseErrorCode::LEC_SYNC_ACTIVATION_TO_DB_FAILED;
    RETURN_VAL_IF_TRUE(new_points_records.size() == 0, PointsRecordVec());

    auto points_records = this->get_points_records();

    // 如果服务号在数据库中已经存在或者导入的服务号已过期，拒绝合并
    for (auto new_iter = new_points_records.begin(); new_iter != new_points_records.end(); new_iter++)
    {
        if (Base::TimeUtils::is_expired((*new_iter)->expiry_date))
        {
            error_code = LicenseErrorCode::LEC_SERVER_IMPORT_POINTS_STATUS_EXPIRED;
            return PointsRecordVec();
        }
        for (auto iter = points_records.begin(); iter != points_records.end(); iter++)
        {
            if ((*new_iter)->uid == (*iter)->uid)
            {
                error_code = LicenseErrorCode::LEC_SERVER_IMPORT_POINTS_ALREADY_EXISTS;
                return PointsRecordVec();
            }
        }
    }

    // 开始合并操作
    for (auto new_iter = new_points_records.begin(); new_iter != new_points_records.end(); new_iter++)
    {
        if (!this->add_points_record(*new_iter))
        {
            error_code == LicenseErrorCode::LEC_SYNC_POINTS_TO_DB_FAILED;
            return PointsRecordVec();
        }
    }

    error_code == LicenseErrorCode::LEC_SUCCESS;
    return new_points_records;
}

bool LicenseDatabase::add_points_record(std::shared_ptr<PointsRecord> points_record)
{
    RETURN_VAL_IF_FALSE(points_record, false);

    auto sql = fmt::format(INSERT_POINTS_RECORD,
                           points_record->uid,
                           points_record->contract_no,
                           points_record->subject_name,
                           points_record->customer_info,
                           points_record->activation_object_name,
                           points_record->expiry_date,
                           points_record->used_on_virtual_machine,
                           points_record->keep_alive_time,
                           points_record->enable_recycling,
                           points_record->points_activation_code,
                           points_record->install_type,
                           points_record->update_mode,
                           points_record->service_mode,
                           points_record->sum_points,
                           points_record->remaining_points,
                           points_record->activation_object_note,
                           points_record->warranty_days,
                           points_record->compute_expiry_day_mode,
                           points_record->warranty_years,
                           points_record->warranty_months,
                           points_record->customization_customer);

    std::string error;
    if (!this->exec_sql(sql, error))
    {
        KLOG_WARNING("%s", error.c_str());
        return false;
    }
    return true;
}

LicenseErrorCode LicenseDatabase::delete_points_record(std::string service_uid)
{
    auto activation_record = get_activation_records(service_uid);
    for (auto iter = activation_record.begin(); iter != activation_record.end(); iter++)
    {
        if (!this->delete_activation_record((*iter)->machine_code))
        {
            KLOG_WARNING("delete activation record failed machine code: %s.", (*iter)->machine_code.c_str());
        }
    }

    auto sql = fmt::format(DELETE_POINTS_RECORD, service_uid);
    std::string error;
    if (!this->exec_sql(sql, error))
    {
        KLOG_WARNING(error.c_str());
        return LicenseErrorCode::LEC_SYNC_POINTS_TO_DB_FAILED;
    }

    return LicenseErrorCode::LEC_SUCCESS;
}

PointsRecordVec LicenseDatabase::get_points_records_by_file(std::string file_name)
{
    sqlite3 *update_file_handle;
    if (sqlite3_open(file_name.c_str(), &update_file_handle) != SQLITE_OK)
    {
        KLOG_WARNING("%s, db_path: %s", sqlite3_errmsg(update_file_handle), file_name.c_str());
        return PointsRecordVec();
    }

    if (sqlite3_key(update_file_handle, DB_KEY, strlen(DB_KEY)) != SQLITE_OK)
    {
        KLOG_WARNING("Failed to decrypt db.");
        return PointsRecordVec();
    }

    char *sqlite3_error = NULL;
    PointsRecordVec points_records;
    int ret = sqlite3_exec(update_file_handle,
                           std::string(SELECT_ALL_POINTS_INFO).c_str(),
                           &LicenseDatabase::parse_points_record,
                           &points_records, &sqlite3_error);
    if (ret != SQLITE_OK)
    {
        KLOG_WARNING("Failed to exec sql {0}. error: {1}.", std::string(SELECT_ALL_POINTS_INFO).c_str(), POINTER_TO_STRING(sqlite3_error).c_str());
        sqlite3_free(sqlite3_error);
        return PointsRecordVec();
    }
    KLOG_DEBUG("Success to exec sql: %s.", std::string(SELECT_ALL_POINTS_INFO).c_str());
    sqlite3_close(update_file_handle);

    return points_records;
}

void LicenseDatabase::process_column_compatible(const std::string &table_name,
                                                const std::string &column_name,
                                                const std::string &value_type,
                                                const std::string &default_value)
{
    std::string error;
    bool is_exits = false;

    this->exec_sql(IS_COLUMN_EXITS(table_name, column_name), error, &LicenseDatabase::check_sql_return, &is_exits);
    if (!is_exits)
    {
        KLOG_INFO("Table %s not have column: %s column will be create", table_name.c_str(), column_name.c_str());
        // 不存在，创建
        if (!this->exec_sql(ADD_COLUMN(table_name, column_name, value_type, default_value), error))
        {
            KLOG_FATAL("Table %s not have column: %s ,create column failed!", table_name.c_str(), column_name.c_str());
        }
    }
}

}  // namespace Server
