/**
 * @file          /ks-license/src/server/license-database.h
 * @brief
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */

#pragma once

#include <sqlcipher/sqlite3.h>
#include <map>
#include <vector>
#include "lib/base/base.h"

namespace Server
{
struct ActivationRecord
{
    std::string service_uid;
    std::string activation_object_name;
    std::string machine_code;
    std::string activation_code;
    std::string expiry_date;
    std::string activation_time;
    uint32_t keep_alive_timestamp;
    std::string client_ip;
};

typedef std::vector<std::shared_ptr<ActivationRecord>> ActivationRecordVec;

struct PointsRecord
{
    PointsRecord() : used_on_virtual_machine(0),
                     keep_alive_time(300),
                     install_type(0),
                     enable_recycling(false),
                     update_mode(LicenseUpdateMode::LICENSE_UPDATE_MODE_NONE),
                     service_mode(LicenseServiceMode::LICENSE_SERVICE_MODE_NONE),
                     sum_points(0),
                     remaining_points(0),
                     compute_expiry_day_mode(ComputeExpiryDayMode::COMPUTE_EXPIRY_DAY_FIXED),
                     warranty_years(0),
                     warranty_months(0),
                     warranty_days(0){};
    // 唯一标识
    std::string uid;
    // 客户信息
    std::string customer_info;
    // 激活对象名称
    std::string activation_object_name;
    // 激活对象名称解释
    std::string activation_object_note;
    // 质保期
    std::string expiry_date;
    // 只在虚拟机中使用
    int32_t used_on_virtual_machine;
    // 连接保活时间（秒）
    int32_t keep_alive_time;
    // 开启回收激活码
    bool enable_recycling;
    // 点数记录激活码
    std::string points_activation_code;
    // 安装类型
    int32_t install_type;
    // 更新模式
    int32_t update_mode;
    // 人工服务模式(5x8或7x24)
    int32_t service_mode;
    // 总激活点数
    int32_t sum_points;
    // 剩余激活点数
    int32_t remaining_points;
    // 合同号
    std::string contract_no;
    // 项目名称
    std::string subject_name;
    // 计算过期时间方式
    int32_t compute_expiry_day_mode;
    /*
     *   以下变量为质保周期，如果计算过期时间方式选择为按照激活时间计算，
     *   该激活码将按照当前时间加上质保周期,即 质保期 = 当前时间 + 质保年数 + 质保月数 + 质保天数
     */
    // 质保年数
    int32_t warranty_years;
    // 质保月数
    int32_t warranty_months;
    // 质保天数
    int32_t warranty_days;
    // 定制客户
    std::string customization_customer;
};

typedef std::vector<std::shared_ptr<PointsRecord>> PointsRecordVec;

class LicenseDatabase
{
public:
    LicenseDatabase();
    virtual ~LicenseDatabase();

    static std::shared_ptr<LicenseDatabase> get_default();

    void init();

    PointsRecordVec get_points_records();
    // 更新已有的点数记录
    bool update_points_record(std::shared_ptr<PointsRecord> points_record);
    // 更新已有点数记录到期时间
    bool update_points_expiry_day(const std::string &uid, const std::string &expiry_day);
    // 更新已有点数记录周期: 该接口只会更新记录，不会加减操作
    bool update_points_by_period(const std::string &uid, int32_t warranty_days, int32_t warranty_months, int32_t warranty_years);
    // 通过文件方式添加新的点数记录
    PointsRecordVec merge_points_file(const std::string &file_name, LicenseErrorCode &error_code);

    // 获取指定服务号的激活记录
    ActivationRecordVec get_activation_records(const std::string &service_uid);
    // 如果激活记录不存在则添加，否则更新
    bool update_activation_record(std::shared_ptr<ActivationRecord> activation_record);
    // 删除激活记录
    bool delete_activation_record(const std::string &machine_code);

    // 添加新的点数记录
    bool add_points_record(std::shared_ptr<PointsRecord> points_record);

    LicenseErrorCode delete_points_record(std::string service_uid);

private:
    // 加载数据库
    bool
    load(std::string &error);

    PointsRecordVec get_points_records_by_file(std::string file_name);

    static int parse_activation_record(void *user_data, int column_num, char **values, char **names);
    static int parse_points_record(void *user_data, int column_num, char **values, char **names);
    static int check_sql_return(void *user_data, int clumn_num, char **values, char **names);

    bool exec_sql(std::string sql, std::string &error, sqlite3_callback callback = NULL, void *user_data = NULL);
    void process_column_compatible(const std::string &table_name,
                                   const std::string &column_name,
                                   const std::string &value_type,
                                   const std::string &default_value);

private:
    static std::shared_ptr<LicenseDatabase> instance_;

    std::string file_path_;
    sqlite3 *db_handle_;

    std::mutex mutex_;
};

}  // namespace Server
