/**
 * @file          /ks-license/src/server/license-impl.cpp
 * @brief
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */

#include "src/server/license-impl.h"
#include <glib/gi18n.h>
#include <stdio.h>
#include <fstream>
#include "lib/base/license-error.h"

namespace Server
{
// 客户端和服务端时间最大差异（秒）
#define CS_TIMESTAMP_MAX_DIFF (60 * 3)
LicenseImpl::LicenseImpl()
{
    this->points_manager_ = LicensePointsManager::get_instance();
    this->database_ = LicenseDatabase::get_default();
}

LicenseImpl::~LicenseImpl()
{
}

::grpc::Status LicenseImpl::GetActivationCode(::grpc::ServerContext* context,
                                              const ::ActivationReq* request,
                                              ::ActivationRsp* response)
{
    do
    {
        // 时间异常处理
        auto current_timestamp = uint32_t(time(NULL));
        response->set_timestamp(current_timestamp);
        if (request->timestamp() > current_timestamp + CS_TIMESTAMP_MAX_DIFF ||
            current_timestamp > request->timestamp() + CS_TIMESTAMP_MAX_DIFF)
        {
            response->set_error_code(LicenseErrorCode::LEC_SERVER_CS_TIME_DIFF_TOO_MUCH);
            break;
        }

        // 获取服务号
        auto license_points = this->points_manager_->get_license_points(request->service_uid());
        if (!license_points)
        {
            response->set_error_code(LicenseErrorCode::LEC_SERVER_SERVICE_UID_NOT_FOUND);
            break;
        }

        if ((license_points->copy_points_record()->customization_customer != request->customization_customer()) &&
            !license_points->copy_points_record()->customization_customer.empty())
        {
            response->set_error_code(LicenseErrorCode::LEC_CUSTOMIZATION_CUSTOMER_NAME_MISMATCH);
            break;
        }

        if (license_points->copy_points_record()->used_on_virtual_machine &&
            request->machine_type() != LicenseMachineType::LICENSE_MACHINE_TYPE_VIRTUAL)
        {
            response->set_error_code(LicenseErrorCode::LEC_SERVER_MACHINE_TYPE_DISMATCH);
            break;
        }

        if (license_points->copy_points_record()->activation_object_name !=
            request->activation_object_name())
        {
            KLOG_WARNING("The service number used active object :%s , but %s has been requied",
                         license_points->copy_points_record()->activation_object_name.c_str(),
                         request->activation_object_name().c_str());
            response->set_error_code(LicenseErrorCode::LEC_SERVER_ACTIVATION_OBJECT_MISMATCH);
            break;
        }

        std::string client_ip = "0.0.0.0";
        auto peer_fields = Base::StrUtils::split_with_char(context->peer(), ':');
        if (peer_fields.size() >= 2)
        {
            client_ip = peer_fields[1];
        }
        else
        {
            KLOG_WARNING("Unkown peer format: %s.", context->peer().c_str());
        }

        std::string activation_code;
        auto error_code = license_points->generate_activation_code(request->machine_code(), client_ip, activation_code);
        if (error_code != LicenseErrorCode::LEC_SUCCESS)
        {
            response->set_error_code(error_code);
        }
        else
        {
            response->set_activation_code(activation_code);
            response->set_keep_alive_time(license_points->copy_points_record()->keep_alive_time);
        }

    } while (0);

    return ::grpc::Status::OK;
}

::grpc::Status LicenseImpl::HeartBeat(::grpc::ServerContext* context,
                                      const ::HeartBeatReq* request,
                                      ::HeartBeatRsp* response)
{
    do
    {
        auto current_timestamp = uint32_t(time(NULL));

        auto license_points = this->points_manager_->get_license_points(request->service_uid());
        if (!license_points)
        {
            response->set_error_code(LicenseErrorCode::LEC_SERVER_SERVICE_UID_NOT_FOUND);
            break;
        }

        auto error_code = license_points->update_keep_alive(request->machine_code());
        if (error_code != LicenseErrorCode::LEC_SUCCESS)
        {
            response->set_error_code(error_code);
            break;
        }
        response->set_keep_alive_time(license_points->copy_points_record()->keep_alive_time);
        response->set_timestamp(current_timestamp);

    } while (0);

    return ::grpc::Status::OK;
}

::grpc::Status LicenseImpl::GetPointsRecords(::grpc::ServerContext* context,
                                             const ::PointsRecordsReq* request,
                                             ::PointsRecordsRsp* response)
{
    auto license_pointss = this->points_manager_->get_license_pointss();
    for (auto iter = license_pointss.begin(); iter != license_pointss.end(); iter++)
    {
        auto point = response->add_points();
        auto points_record = (*iter)->copy_points_record();
        point->set_uid(points_record->uid);
        point->set_customer_info(points_record->customer_info);
        point->set_activation_object_name(points_record->activation_object_name);
        point->set_activation_object_notes(points_record->activation_object_note);
        point->set_expiry_date(points_record->expiry_date);
        point->set_install_type(points_record->install_type);
        point->set_update_mode(points_record->update_mode);
        point->set_service_mode(points_record->service_mode);
        point->set_sum_points(points_record->sum_points);
        point->set_contract_no(points_record->contract_no);
        point->set_subject_name(points_record->subject_name);
        point->set_remaining_points(points_record->remaining_points);
        point->set_enable_recycling(points_record->enable_recycling);
        point->set_compute_expiry_day_mode(points_record->compute_expiry_day_mode);
        point->set_warranty_days(points_record->warranty_days);
        point->set_warranty_months(points_record->warranty_months);
        point->set_warranty_years(points_record->warranty_years);
        point->set_customization_customer(points_record->customization_customer);
    }

    return ::grpc::Status::OK;
}

::grpc::Status LicenseImpl::GetActivationRecords(::grpc::ServerContext* context,
                                                 const ::ActivationRecordReq* request,
                                                 ::ActivationRecordsRsp* response)
{
    auto license_points = this->points_manager_->get_license_points(request->service_uid());

    if (license_points)
    {
        auto points_records = license_points->copy_points_record();
        auto activation_records = license_points->copy_activation_records();

        for (auto iter = activation_records.begin(); iter != activation_records.end(); iter++)
        {
            auto activation_record_pb = response->add_activation_records();
            this->fill_activation_records(activation_record_pb, (*iter));
        }
    }
    return ::grpc::Status::OK;
}

::grpc::Status LicenseImpl::ImportPointsRecords(::grpc::ServerContext* context,
                                                const ::ImportPointsReq* request,
                                                ::ImportPointsRsp* response)
{
    auto points_records_bytes = request->points_records_bytes();
    auto filename = fmt::format("{0}/import_points_tmp_{1}.db", KL_INSTALL_DATADIR, rand());
    std::ofstream fout(filename, std::ios::out | std::ios::binary);
    do
    {
        if (!fout)
        {
            response->set_error(Base::LicenseError::get_error_desc(LicenseErrorCode::LEC_COMMON_IO_ERROR));
            break;
        }
        fout.write(points_records_bytes.c_str(), points_records_bytes.length());
        fout.flush();

        LicenseErrorCode error_code;
        if (!this->points_manager_->import_points_file(filename, error_code))
        {
            response->set_error(Base::LicenseError::get_error_desc(error_code));
            KLOG_WARNING(Base::LicenseError::get_error_desc(error_code).c_str());
            break;
        }
    } while (0);

    if (fout)
    {
        fout.close();
        remove(filename.c_str());
    }

    return ::grpc::Status::OK;
}

::grpc::Status LicenseImpl::GetServerInfo(::grpc::ServerContext* context,
                                          const ::ServerInfoReq* request,
                                          ::ServerInfoRsp* response)
{
    struct ::ifaddrs *ifaddr, *ifa;

    if (getifaddrs(&ifaddr) == -1)
    {
        KLOG_WARNING("getifaddrs:%s", ::strerror(errno));
    }

    for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next)
    {
        if (ifa->ifa_addr == NULL)
            continue;

        int family = ifa->ifa_addr->sa_family;

        if (family == AF_INET)
        {
            auto ip = response->add_ip();
            *ip = std::string(::inet_ntoa(((struct sockaddr_in*)ifa->ifa_addr)->sin_addr));
        }
    }

    freeifaddrs(ifaddr);

    response->set_port(DEFAULT_LISTEN_PORT);
    response->set_machine_code(this->get_machine_code());

    return ::grpc::Status::OK;
}

::grpc::Status LicenseImpl::GetActivationRecordsByCond(::grpc::ServerContext* context,
                                                       const ::ActivationRecordsByCondReq* request,
                                                       ::ActivationRecordsByCondRsp* response)
{
    std::unique_lock<std::mutex> lock(this->mutex_);
    if (!request->use_cached())
    {
        this->update_activations_cache(request->keywords());
    }

    int max_page = ceil((double)this->activations_cache_.size() / (double)request->page_size());
    if (this->activations_cache_.size() == 0)
    {
        max_page = 1;
    }

    if (request->end_page_no() - request->start_page_no() < 0 ||
        request->end_page_no() > max_page ||
        request->start_page_no() > max_page)
    {
        // 请求页码错误
        response->set_error(Base::LicenseError::get_error_desc(LicenseErrorCode::LEC_SERVER_REQUEST_PAGE_NO_ERROR));
        return ::grpc::Status::OK;
    }

    for (int i = request->page_size() * (request->start_page_no() - 1);
         i < request->page_size() * (request->end_page_no());
         i++)
    {
        if (i >= this->activations_cache_.size())
        {
            break;
        }
        auto activation_record_pb = response->add_query_pages_activation_records();
        this->fill_activation_records(activation_record_pb, this->activations_cache_[i]);
    }

    response->set_all_pages_activation_records(this->activations_cache_.size());
    return ::grpc::Status::OK;
}

::grpc::Status LicenseImpl::UpdatePointsRecords(::grpc::ServerContext* context, const ::UpdatePointsReq* request, ::UpdatePointsRsp* response)
{
    auto is_update = request->is_update();
    auto points = request->points_record();

    auto points_record = std::make_shared<PointsRecord>();

    points_record->uid = points.uid();
    points_record->customer_info = points.customer_info();
    points_record->activation_object_name = points.activation_object_name();
    points_record->expiry_date = points.expiry_date();
    points_record->used_on_virtual_machine = points.used_on_virtual_machine();
    points_record->keep_alive_time = points.keep_alive_time();
    points_record->enable_recycling = points.enable_recycling();
    points_record->points_activation_code = points.points_activation_code();
    points_record->update_mode = points.update_mode();
    points_record->service_mode = points.service_mode();
    points_record->sum_points = points.sum_points();
    points_record->remaining_points = points.remaining_points();
    points_record->contract_no = points.contract_no();
    points_record->subject_name = points.subject_name();
    points_record->install_type = points.install_type();
    points_record->warranty_days = points.warranty_days();
    points_record->compute_expiry_day_mode = points.compute_expiry_day_mode();
    points_record->warranty_months = points.warranty_months();
    points_record->warranty_years = points.warranty_years();
    points_record->customization_customer = points.customization_customer();

    if (is_update)
    {
        if (points_record->compute_expiry_day_mode == ComputeExpiryDayMode::COMPUTE_EXPIRY_DAY_BY_ACTIVATE)
        {
            response->set_error_code(this->points_manager_->update_points(points_record->uid,
                                                                          points_record->warranty_days,
                                                                          points_record->warranty_months,
                                                                          points_record->warranty_years));
        }
        else
        {
            response->set_error_code(this->points_manager_->update_points(points_record->uid,
                                                                          points_record->expiry_date));
        }
    }
    else
    {
        response->set_error_code(this->points_manager_->insert_points(points_record));
    }

    return ::grpc::Status::OK;
}

::grpc::Status LicenseImpl::DeletePointsRecords(::grpc::ServerContext* context, const ::DeletePointsReq* request, ::DeletePointsRsp* response)
{
    auto service_uid = request->service_uid();
    auto points_record = this->points_manager_->get_license_points(service_uid);
    if (!points_record)
    {
        response->set_error_code(LicenseErrorCode::LEC_SERVER_SERVICE_UID_NOT_FOUND);
        return ::grpc::Status::OK;
    }

    response->set_error_code(this->points_manager_->delete_points(service_uid));

    return ::grpc::Status::OK;
}

void LicenseImpl::fill_activation_records(::ActivationRecord* activation_record_pb, std::shared_ptr<ActivationRecord> activation_record)
{
    int keep_alive_time = 0;

    activation_record_pb->set_service_uid(activation_record->service_uid);
    activation_record_pb->set_activation_object_name(activation_record->activation_object_name);
    activation_record_pb->set_machine_code(activation_record->machine_code);
    activation_record_pb->set_activation_code(activation_record->activation_code);
    activation_record_pb->set_expiry_date(activation_record->expiry_date);
    activation_record_pb->set_activation_time(activation_record->activation_time);

    auto license_points = this->points_manager_->get_license_points(activation_record->service_uid);
    if (license_points)
    {
        keep_alive_time = license_points->copy_points_record()->keep_alive_time;
    }

    if (!keep_alive_time)
    {
        keep_alive_time = HEART_BEAT_INTERVAL * 5;
    }

    // 设定3次心跳时间内有收到一次心跳都算在线
    if (activation_record->keep_alive_timestamp + keep_alive_time > time(NULL))
    {
        activation_record_pb->set_client_connection_status(ClientConnectionStatus::CLIENT_CONNECTION_STATUS_ONLINE);
    }
    else
    {
        activation_record_pb->set_client_connection_status(ClientConnectionStatus::CLIENT_CONNECTION_STATUS_OFFLINE);
    }
    activation_record_pb->set_client_ip(activation_record->client_ip);
}

void LicenseImpl::update_activations_cache(const std::string& keywords)
{
    this->activations_cache_.clear();

    // 遍历筛选所有的激活信息
    auto license_pointss = this->points_manager_->get_license_pointss();
    for (int i = 0; i < license_pointss.size(); i++)
    {
        auto activation_records = license_pointss[i]->copy_activation_records();
        for (auto iter = activation_records.begin(); iter != activation_records.end(); iter++)
        {
            // 适配筛选条件
            if (!keywords.empty())
            {
                std::map<std::string, std::string> iter_data_map{{"service_uid", (*iter)->service_uid},
                                                                 {"activation_object_name", (*iter)->activation_object_name},
                                                                 {"machine_code", (*iter)->machine_code},
                                                                 {"activation_code", (*iter)->activation_code},
                                                                 {"expiry_date", (*iter)->expiry_date},
                                                                 {"activation_time", (*iter)->activation_time},
                                                                 {"keep_alive_timestamp", std::to_string((*iter)->keep_alive_timestamp)},
                                                                 {"client_ip", (*iter)->client_ip}};
                // 精确搜索
                if (keywords[0] == '$' && keywords.find('=') != std::string::npos)
                {
                    if (this->accurate_search_activation_info(iter_data_map, keywords))
                    {
                        this->activations_cache_.push_back(*iter);
                    }
                }
                // 模糊搜索
                else
                {
                    if (this->fuzzy_search_activation_info(iter_data_map, keywords))
                    {
                        this->activations_cache_.push_back(*iter);
                    }
                }
            }
            else
            {
                this->activations_cache_.insert(std::end(this->activations_cache_), (*iter));
            }
        }
    }
}

bool LicenseImpl::fuzzy_search_activation_info(const std::map<std::string, std::string>& activation_info, const std::string& keywords)
{
    for (auto iter = activation_info.begin(); iter != activation_info.end(); iter++)
    {
        auto result = (*iter).second.find(keywords);
        if (result != std::string::npos)
        {
            return true;
        }
    }

    return false;
}

bool LicenseImpl::accurate_search_activation_info(const std::map<std::string, std::string>& activation_info, const std::string& keywords)
{
    auto sub_str = keywords.substr(1);
    auto result = sub_str.find('=');
    auto key = sub_str.substr(0, result);
    auto value = sub_str.substr(result + 1);

    auto iter = activation_info.find(key);
    if (iter != activation_info.end())
    {
        return (value == iter->second);
    }
    else
    {
        return false;
    }
}

std::string LicenseImpl::get_machine_code()
{
    auto command_line = fmt::format("/usr/bin/ksl-register -s -j");
    std::string standard_output;
    std::string standard_error;
    Glib::spawn_command_line_sync(command_line, &standard_output, &standard_error);

    if (standard_error.length() > 0)
    {
        KLOG_WARNING("%s", standard_error.c_str());
        return std::string();
    }

    auto license_value = Base::StrUtils::str2json(standard_output);
    return license_value[LICENSE_JK_MACHINE_CODE].asString();
}

}  // namespace Server
