/**
 * @file          /ks-license/src/server/license-points-manager.cpp
 * @brief
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */

#include "src/server/license-points-manager.h"
#include <iostream>
#include "src/server/license-points.h"

namespace Server
{
LicensePointsManager::LicensePointsManager()
{
    this->database_ = LicenseDatabase::get_default();
}

LicensePointsManager::~LicensePointsManager()
{
}

LicensePointsManager *LicensePointsManager::instance_ = NULL;
void LicensePointsManager::global_init()
{
    instance_ = new LicensePointsManager();
    instance_->init();
}

LicensePointsVec LicensePointsManager::get_license_pointss()
{
    LicensePointsVec points_records;
    {
        std::unique_lock<std::mutex> lock(this->mutex_);
        for (auto iter = this->points_records_.begin(); iter != this->points_records_.end(); ++iter)
        {
            points_records.push_back(iter->second);
        }
    }
    return points_records;
}

std::shared_ptr<LicensePoints> LicensePointsManager::get_license_points(const std::string &service_uid)
{
    std::unique_lock<std::mutex> lock(this->mutex_);
    return Base::MapHelper::get_value(this->points_records_, service_uid);
}

bool LicensePointsManager::import_points_file(const std::string &file_name, LicenseErrorCode &error_code)
{
    auto points_records = this->database_->merge_points_file(file_name, error_code);
    RETURN_VAL_IF_TRUE(points_records.size() == 0, false);

    for (uint32_t i = 0; i < points_records.size(); ++i)
    {
        auto &points_record = points_records[i];
        auto license_points = std::make_shared<LicensePoints>(points_record);

        if (license_points->init())
        {
            this->add_license_points(license_points);
        }
        else
        {
            error_code = LicenseErrorCode::LEC_SERVER_INIT_POINTS_FAILED;
            return false;
        }
    }
    return true;
}

LicenseErrorCode LicensePointsManager::insert_points(std::shared_ptr<PointsRecord> points_record)
{
    if (this->get_license_points(points_record->uid))
    {
        return LicenseErrorCode::LEC_SERVER_IMPORT_POINTS_ALREADY_EXISTS;
    }

    return this->add_new_license_points(points_record);
}

LicenseErrorCode LicensePointsManager::update_points(const std::string &service_uid, const std::string &expiry_day)
{
    if (!this->database_->update_points_expiry_day(service_uid, expiry_day))
    {
        return LicenseErrorCode::LEC_SYNC_POINTS_TO_DB_FAILED;
    }

    auto license_points = this->get_license_points(service_uid);
    RETURN_VAL_IF_FALSE(license_points, LicenseErrorCode::LEC_SERVER_SERVICE_UID_NOT_FOUND);

    auto copy_points_record = license_points->copy_points_record();
    copy_points_record->expiry_date = expiry_day;
    license_points->set_points_record(copy_points_record);

    GDate expiry_g_date = Base::TimeUtils::timestamp2gdate(Base::TimeUtils::str2timestamp(expiry_day.c_str()));

    if (!license_points->activations_update_expiry_time(expiry_g_date))
    {
        return LicenseErrorCode::LEC_SERVER_ACTIVATION_RECORD_UPDATE_FAILED;
    }

    return LicenseErrorCode::LEC_SUCCESS;
}

LicenseErrorCode LicensePointsManager::update_points(const std::string &service_uid, int32_t warranty_days, int32_t warranty_months, int32_t warranty_years)
{
    auto license_points = this->get_license_points(service_uid);
    RETURN_VAL_IF_FALSE(license_points, LicenseErrorCode::LEC_SERVER_SERVICE_UID_NOT_FOUND);
    auto copy_points_record = license_points->copy_points_record();
    copy_points_record->warranty_years += warranty_years;
    copy_points_record->warranty_months += warranty_months;
    copy_points_record->warranty_days += warranty_days;

    if (!this->database_->update_points_by_period(service_uid,
                                                  copy_points_record->warranty_days,
                                                  copy_points_record->warranty_months,
                                                  copy_points_record->warranty_years))
    {
        return LicenseErrorCode::LEC_SYNC_POINTS_TO_DB_FAILED;
    }

    license_points->set_points_record(copy_points_record);
    if (!license_points->activations_add_period(warranty_days, warranty_months, warranty_years))
    {
        return LicenseErrorCode::LEC_SERVER_ACTIVATION_RECORD_UPDATE_FAILED;
    }

    return LicenseErrorCode::LEC_SUCCESS;
}

LicenseErrorCode LicensePointsManager::delete_points(const std::string &service_uid)
{
    if (!this->points_records_.erase(service_uid))
    {
        return LicenseErrorCode::LEC_SERVER_SERVICE_UID_NOT_FOUND;
    }

    return this->database_->delete_points_record(service_uid);
}

void LicensePointsManager::init()
{
    this->database_->init();
    this->load_points();
}

void LicensePointsManager::load_points()
{
    KLOG_PROFILE("");

    auto points_records = this->database_->get_points_records();
    for (uint32_t i = 0; i < points_records.size(); ++i)
    {
        auto &points_record = points_records[i];
        auto license_points = std::make_shared<LicensePoints>(points_record);

        if (license_points->init())
        {
            this->add_license_points(license_points);
        }
    }
}

void LicensePointsManager::add_license_points(std::shared_ptr<LicensePoints> license_points)
{
    RETURN_IF_FALSE(license_points);
    {
        std::unique_lock<std::mutex> lock(this->mutex_);
        if (this->points_records_.find(license_points->get_uid()) == this->points_records_.end())
        {
            this->points_records_[license_points->get_uid()] = license_points;
        }
        else
        {
            KLOG_WARNING("The id %s is already exist", license_points->get_uid().c_str());
        }
    }
}

LicenseErrorCode LicensePointsManager::add_new_license_points(std::shared_ptr<PointsRecord> points_record)
{
    auto license_points = std::make_shared<LicensePoints>(points_record);

    if (license_points->init())
    {
        this->add_license_points(license_points);

        auto res = this->database_->add_points_record(points_record);
        if (!res)
        {
            return LicenseErrorCode::LEC_SYNC_POINTS_TO_DB_FAILED;
        }
        return LicenseErrorCode::LEC_SUCCESS;
    }

    return LicenseErrorCode::LEC_SERVER_INIT_POINTS_FAILED;
}

}  // namespace Server
