/**
 * @file          /ks-license/src/server/license-points.cpp
 * @brief
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */

#include "src/server/license-points.h"
#include <dbus-c++/glib-integration.h>
#include <glib/gi18n.h>
#include "lib/base/base.h"
#include "src/server/thread-pool.h"

namespace Server
{
// 虚拟ukey设备每次授权的时间
#define UKEY_VIRTUAL_MAX_AUTHORIZED_DAYS 7

LicensePoints::LicensePoints(std::shared_ptr<PointsRecord> points_record) : points_record_(points_record),
                                                                            is_dirty_(false)
{
}

bool LicensePoints::init()
{
    auto ukey_database = LicenseDatabase::get_default();
    auto activation_records = ukey_database->get_activation_records(this->points_record_->uid);
    for (uint32_t i = 0; i < activation_records.size(); ++i)
    {
        auto &activation_record = activation_records[i];
        this->activation_records_[activation_record->machine_code] = activation_record;
    }

    // 校验激活码是否合法，只有激活码和当前机器匹配才能使用该点数记录
    try
    {
        auto command_line = fmt::format("/usr/bin/ksl-register -a {0} -o {1} -r -s -j --wait-service", points_record_->points_activation_code, LICENSE_OBJECT_VUKEY_NAME);
        std::string standard_output;
        std::string standard_error;
        Glib::spawn_command_line_sync(command_line, &standard_output, &standard_error);

        if (standard_error.length() > 0)
        {
            KLOG_WARNING("%s", standard_error.c_str());
            return false;
        }

        auto license_value = Base::StrUtils::str2json(standard_output);

        if (license_value[LICENSE_JK_ACTIVATION_STATUS].asInt() != LicenseActivationStatus::LAS_ACTIVATED ||
            !(license_value[LICENSE_JK_SERVICE_STATUS].asInt() == LicenseServiceStatus::LICENSE_SERVICE_STATUS_UNEXPIRED ||
              license_value[LICENSE_JK_SERVICE_STATUS].asInt() == LicenseServiceStatus::LICENSE_SERVICE_STATUS_EXPIRED))
        {
            KLOG_WARNING("The points record is unactivated.");
            return false;
        }
        this->print_points();
    }
    catch (const Glib::Error &e)
    {
        KLOG_WARNING("%s", e.what().c_str());
        return false;
    }

    return true;
}

std::shared_ptr<PointsRecord> LicensePoints::copy_points_record()
{
    std::unique_lock<std::recursive_mutex> lock(this->mutex_);
    auto points_record = std::make_shared<PointsRecord>(*(this->points_record_));
    return points_record;
};

bool LicensePoints::set_points_record(std::shared_ptr<PointsRecord> points_record)
{
    if (points_record->uid != this->points_record_->uid)
    {
        return false;
    }
    this->points_record_ = points_record;
    return true;
}

ActivationRecordVec LicensePoints::copy_activation_records()
{
    ActivationRecordVec activation_records;

    this->process_activation_recycling();
    {
        std::unique_lock<std::recursive_mutex> lock(this->mutex_);
        for (auto iter = this->activation_records_.begin(); iter != this->activation_records_.end(); ++iter)
        {
            activation_records.push_back(std::make_shared<ActivationRecord>(*(iter->second)));
        }
    }

    return activation_records;
}

LicenseErrorCode LicensePoints::generate_activation_code(const std::string &machine_code_str,
                                                         const std::string &client_ip,
                                                         std::string &out_license_code)
{
    std::string machine_code;
    unsigned char activation_code[LICENSE_CODE_LENGTH] = {0};
    unsigned char expiry_date_buf[OLD_EXPIRY_DATE_LENGTH];
    std::string encrypted_activation_code;
    int32_t decreasing_point = 1;
    std::string expiry_date_str;
    // 同步已激活信息
    auto current_timestamp = uint32_t(time(NULL));

    std::unique_lock<std::recursive_mutex> lock(this->mutex_);

    RETURN_VAL_IF_ERROR(Base::LicenseUtils::unformat_machine_code(machine_code_str, machine_code));

    if (this->points_record_->install_type != LicenseInstallType::LICENSE_INSTALL_TYPE_UNLIMIT &&
        this->points_record_->install_type != LicenseInstallType(machine_code[0] & 0xF))
    {
        return LicenseErrorCode::MACHINE_CODE_UNMATCH_INSTALL_TYPE;
    }

    // 已经激活过的机器码不消耗点数，只更新激活记录
    if (this->exist_machine_code(machine_code_str))
    {
        decreasing_point = 0;
        KLOG_DEBUG("The machine is already registered.");
        out_license_code = this->get_activation_record(machine_code_str)->activation_code;
        expiry_date_str = this->get_activation_record(machine_code_str)->expiry_date;
    }
    else
    {
        GDate expiry_date;
        if (points_record_->compute_expiry_day_mode == ComputeExpiryDayMode::COMPUTE_EXPIRY_DAY_BY_ACTIVATE)
        {
            // 生成到期时间
            expiry_date = this->generate_expiry_time_by_period(this->points_record_->warranty_days,
                                                               this->points_record_->warranty_months,
                                                               this->points_record_->warranty_years);
        }
        else
        {
            g_date_set_parse(&expiry_date, this->points_record_->expiry_date.c_str());
            if (!g_date_valid(&expiry_date))
            {
                return LicenseErrorCode::LEC_UKEY_EXPIRED_DATE_INVALIDE;
            }
        }
        expiry_date_str = Base::TimeUtils::gdate2str(&expiry_date);

        memcpy(activation_code, machine_code.data(), MACHINE_CODE_LENGTH);
        if (this->points_record_->customer_info.length() > 0)
        {
            // 旧的激活码格式：机器码+质保期+客户信息
            RETURN_VAL_IF_ERROR(Base::TimeUtils::gdate2arr(&expiry_date, expiry_date_buf));
            memcpy(activation_code + MACHINE_CODE_LENGTH, expiry_date_buf, OLD_EXPIRY_DATE_LENGTH);
            memcpy(activation_code + MACHINE_CODE_LENGTH + OLD_EXPIRY_DATE_LENGTH, this->points_record_->customer_info.c_str(), CUSTOMER_CODE_LENGTH);
        }
        else
        {
            // 新的激活码格式：机器码+质保期+更新模式+服务模式
            // 第一个字节最高位标记为1表示新的格式
            activation_code[0] |= ACTIVATION_CODE_VERSION_MASK;
            // 由于过期日期是按天为单位的，所以时间戳设置为当天的最后一秒
            uint32_t expiry_timestamp = uint32_t(Base::TimeUtils::gdate2timestamp(&expiry_date) + DAY_IN_SECONDS - 1);
            memcpy(activation_code + MACHINE_CODE_LENGTH, &expiry_timestamp, EXPIRY_DATE_LENGTH);

            activation_code[MACHINE_CODE_LENGTH + EXPIRY_DATE_LENGTH] |= (this->points_record_->update_mode << LICENSE_UPDATE_MODE_BIT_POS);
            activation_code[MACHINE_CODE_LENGTH + EXPIRY_DATE_LENGTH] |= (this->points_record_->service_mode << LICENSE_SERVICE_MODE_BIT_POS);
            activation_code[MACHINE_CODE_LENGTH + EXPIRY_DATE_LENGTH] |= (LICENSE_ACTIVATION_TYPE_ONLINE << LICENSE_ACTIVATION_TYPE_BIT_POS);
        }

        auto key = Base::LicenseUtils::generate_key(this->points_record_->activation_object_name);
        if (!Base::CryptoHelper::des_encrypt(std::string((char *)activation_code, LICENSE_CODE_LENGTH), encrypted_activation_code, key))
        {
            return LicenseErrorCode::LEC_ENCODE_GENERATE_LICENSE_FAILED;
        }

        RETURN_VAL_IF_FALSE(encrypted_activation_code.length() == LICENSE_CODE_LENGTH, LicenseErrorCode::LEC_GENERATE_LICENSE_LENGTH_INALID);
        // 点数已经用完则尝试回收过期机器
        if (decreasing_point > 0 && this->points_record_->remaining_points <= 0)
        {
            this->process_activation_recycling();
            RETURN_VAL_IF_TRUE(this->points_record_->remaining_points <= 0, LicenseErrorCode::LEC_SERVER_NO_REMAINING_POINTS);
        }
        out_license_code = Base::LicenseUtils::format_activation_code(encrypted_activation_code);
    }

    this->points_record_->remaining_points -= decreasing_point;
    auto activation_record = std::make_shared<ActivationRecord>(ActivationRecord{this->points_record_->uid,
                                                                                 this->points_record_->activation_object_name,
                                                                                 machine_code_str,
                                                                                 out_license_code,
                                                                                 expiry_date_str,
                                                                                 Base::TimeUtils::timestamp2str(current_timestamp),
                                                                                 current_timestamp,
                                                                                 client_ip});

    this->activation_records_[activation_record->machine_code] = activation_record;

    this->cache_dirty_records(DirtyActivationRecord::OperationType::OPERATION_TYPE_ADD, ActivationRecordVec{activation_record});

    this->print_points();
    return LicenseErrorCode::LEC_SUCCESS;
}

LicenseErrorCode LicensePoints::update_activation_code(const std::string &machine_code_str,
                                                       GDate expiry_date,
                                                       std::string &out_license_code)
{
    RETURN_VAL_IF_FALSE(this->exist_machine_code(machine_code_str), LicenseErrorCode::LEC_SERVER_NOT_EXIST_MACHINE_CODE);

    std::string machine_code;
    unsigned char activation_code[LICENSE_CODE_LENGTH] = {0};
    unsigned char expiry_date_buf[OLD_EXPIRY_DATE_LENGTH];
    std::string encrypted_activation_code;
    // 同步已激活信息
    auto current_timestamp = uint32_t(time(NULL));

    RETURN_VAL_IF_ERROR(Base::LicenseUtils::unformat_machine_code(machine_code_str, machine_code));

    if (this->points_record_->install_type != LicenseInstallType::LICENSE_INSTALL_TYPE_UNLIMIT &&
        this->points_record_->install_type != LicenseInstallType(machine_code[0] & 0xF))
    {
        return LicenseErrorCode::MACHINE_CODE_UNMATCH_INSTALL_TYPE;
    }

    memcpy(activation_code, machine_code.data(), MACHINE_CODE_LENGTH);
    if (this->points_record_->customer_info.length() > 0)
    {
        // 旧的激活码格式：机器码+质保期+客户信息
        RETURN_VAL_IF_ERROR(Base::TimeUtils::gdate2arr(&expiry_date, expiry_date_buf));
        memcpy(activation_code + MACHINE_CODE_LENGTH, expiry_date_buf, OLD_EXPIRY_DATE_LENGTH);
        memcpy(activation_code + MACHINE_CODE_LENGTH + OLD_EXPIRY_DATE_LENGTH, this->points_record_->customer_info.c_str(), CUSTOMER_CODE_LENGTH);
    }
    else
    {
        // 新的激活码格式：机器码+质保期+更新模式+服务模式
        // 第一个字节最高位标记为1表示新的格式
        activation_code[0] |= ACTIVATION_CODE_VERSION_MASK;
        // 由于过期日期是按天为单位的，所以时间戳设置为当天的最后一秒
        uint32_t expiry_timestamp = uint32_t(Base::TimeUtils::gdate2timestamp(&expiry_date) + DAY_IN_SECONDS - 1);
        memcpy(activation_code + MACHINE_CODE_LENGTH, &expiry_timestamp, EXPIRY_DATE_LENGTH);

        activation_code[MACHINE_CODE_LENGTH + EXPIRY_DATE_LENGTH] |= (this->points_record_->update_mode << LICENSE_UPDATE_MODE_BIT_POS);
        activation_code[MACHINE_CODE_LENGTH + EXPIRY_DATE_LENGTH] |= (this->points_record_->service_mode << LICENSE_SERVICE_MODE_BIT_POS);
        activation_code[MACHINE_CODE_LENGTH + EXPIRY_DATE_LENGTH] |= (LICENSE_ACTIVATION_TYPE_ONLINE << LICENSE_ACTIVATION_TYPE_BIT_POS);
    }

    auto key = Base::LicenseUtils::generate_key(this->points_record_->activation_object_name);
    if (!Base::CryptoHelper::des_encrypt(std::string((char *)activation_code, LICENSE_CODE_LENGTH), encrypted_activation_code, key))
    {
        return LicenseErrorCode::LEC_ENCODE_GENERATE_LICENSE_FAILED;
    }

    RETURN_VAL_IF_FALSE(encrypted_activation_code.length() == LICENSE_CODE_LENGTH, LicenseErrorCode::LEC_GENERATE_LICENSE_LENGTH_INALID);
    out_license_code = Base::LicenseUtils::format_activation_code(encrypted_activation_code);

    return LicenseErrorCode::LEC_SUCCESS;
}

bool LicensePoints::exist_machine_code(const std::string &machine_code)
{
    std::unique_lock<std::recursive_mutex> lock(this->mutex_);
    auto iter = this->activation_records_.find(machine_code);
    RETURN_VAL_IF_TRUE(iter == this->activation_records_.end(), false);
    return true;
}

LicenseErrorCode LicensePoints::update_keep_alive(const std::string &machine_code)
{
    {
        std::unique_lock<std::recursive_mutex> lock(this->mutex_);
        auto activation_record = this->get_activation_record(machine_code);
        RETURN_VAL_IF_FALSE(activation_record, LicenseErrorCode::LEC_SERVER_ACTIVATION_RECORD_NOT_FOUND);
        activation_record->keep_alive_timestamp = uint32_t(time(NULL));
        this->cache_dirty_records(DirtyActivationRecord::OperationType::OPERATION_TYPE_MOD, ActivationRecordVec{activation_record});
    }
    return LicenseErrorCode::LEC_SUCCESS;
}

ActivationRecordVec LicensePoints::get_activation_records()
{
    ActivationRecordVec activation_records;

    this->process_activation_recycling();
    {
        std::unique_lock<std::recursive_mutex> lock(this->mutex_);
        for (auto iter = this->activation_records_.begin(); iter != this->activation_records_.end(); ++iter)
        {
            activation_records.push_back(iter->second);
        }
    }

    return activation_records;
}

std::shared_ptr<ActivationRecord> LicensePoints::get_activation_record(const std::string &machine_code)
{
    std::unique_lock<std::recursive_mutex> lock(this->mutex_);
    return Base::MapHelper::get_value(this->activation_records_, machine_code);
};

void LicensePoints::print_points()
{
    std::unique_lock<std::recursive_mutex> lock(this->mutex_);

    KLOG_DEBUG("UID:                         %s", this->points_record_->uid.c_str());
    KLOG_DEBUG("ContractNo:                  %s", this->points_record_->contract_no.c_str());
    KLOG_DEBUG("SubjectName:                 %s", this->points_record_->subject_name.c_str());
    KLOG_DEBUG("CustomerInfo:                %s", this->points_record_->customer_info.c_str());
    KLOG_DEBUG("ActivationObjectName:        %s", this->points_record_->activation_object_name.c_str());
    KLOG_DEBUG("ExpredDate:                  %s", this->points_record_->expiry_date.c_str());
    KLOG_DEBUG("Used on virtual machine:     %d", this->points_record_->used_on_virtual_machine);
    KLOG_DEBUG("Keep alive time:             %d", this->points_record_->keep_alive_time);
    KLOG_DEBUG("Enable recycling:            %d", this->points_record_->enable_recycling);
    KLOG_DEBUG("InstallType:                 %d", this->points_record_->install_type);
    KLOG_DEBUG("UpdateMode:                  %d", this->points_record_->update_mode);
    KLOG_DEBUG("ServiceMode:                 %d", this->points_record_->service_mode);
    KLOG_DEBUG("SumPoints:                   %d", this->points_record_->sum_points);
    KLOG_DEBUG("RemainingPoints:             %d", this->points_record_->remaining_points);
    KLOG_DEBUG("ActivationMachineCodes:");
    for (auto iter = this->activation_records_.begin(); iter != this->activation_records_.end(); ++iter)
    {
        KLOG_DEBUG("    %s", (iter->first).c_str());
    }
}

void LicensePoints::process_activation_recycling()
{
    // 扫描需要回收的记录
    ActivationRecordVec recycling_activation_records;

    auto current_timestamp = uint32_t(time(NULL));
    std::unique_lock<std::recursive_mutex> lock(this->mutex_);

    RETURN_IF_TRUE(!this->points_record_->enable_recycling);

    for (auto iter = this->activation_records_.begin(); iter != this->activation_records_.end(); ++iter)
    {
        if (this->points_record_->keep_alive_time > 0 &&
            iter->second->keep_alive_timestamp + this->points_record_->keep_alive_time < current_timestamp)
        {
            KLOG_DEBUG("Recycle machine %s, keep alive timestamp: %u, keep alive time: %u, current timestamp: %u.",
                       iter->second->machine_code.c_str(),
                       iter->second->keep_alive_timestamp,
                       this->points_record_->keep_alive_time,
                       current_timestamp);
            recycling_activation_records.push_back(iter->second);
        }
    }

    // 进行回收操作
    if (recycling_activation_records.size() > 0)
    {
        for (auto iter = recycling_activation_records.begin(); iter != recycling_activation_records.end(); ++iter)
        {
            this->activation_records_.erase((*iter)->machine_code);
        }
        this->points_record_->remaining_points += recycling_activation_records.size();
    }

    if (recycling_activation_records.size() > 0)
    {
        this->cache_dirty_records(DirtyActivationRecord::OperationType::OPERATION_TYPE_DEL, recycling_activation_records);
    }
}

void LicensePoints::cache_dirty_records(DirtyActivationRecord::OperationType operation_type, const ActivationRecordVec &activation_records)
{
    std::unique_lock<std::mutex> dirty_lock(this->dirty_mutex_);
    {
        // 需要拷贝共享指针的数据
        std::unique_lock<std::recursive_mutex> lock(this->mutex_);
        for (auto iter = activation_records.begin(); iter != activation_records.end(); ++iter)
        {
            DirtyActivationRecord dirty_activation_record{operation_type, std::make_shared<ActivationRecord>(*(*iter))};
            this->dirty_points_info_.activation_records[dirty_activation_record.activation_record->machine_code] = dirty_activation_record;
        }
        this->dirty_points_info_.points_record = std::make_shared<PointsRecord>(*this->points_record_);
    }

    if (!this->is_dirty_)
    {
        this->is_dirty_ = true;
        ThreadPool::get_default()->enqueue(std::bind(&LicensePoints::flush_to_database, this));
    }
}

void LicensePoints::flush_to_database()
{
    DirtyPointsInfo dirty_points_info;
    {
        std::unique_lock<std::mutex> lock(this->dirty_mutex_);
        dirty_points_info = std::move(this->dirty_points_info_);
        this->is_dirty_ = false;
    }

    KLOG_WARNING("Flush activation record number: %d.", dirty_points_info.activation_records.size());

    RETURN_IF_TRUE(dirty_points_info.activation_records.size() == 0);

    // 可能有两个线程都会进入flush_to_database，这里需要确保同步到数据库的顺序
    {
        std::unique_lock<std::mutex> lock(this->flush_db_mutex_);
        auto database = LicenseDatabase::get_default();
        database->update_points_record(dirty_points_info.points_record);

        for (auto iter = dirty_points_info.activation_records.begin(); iter != dirty_points_info.activation_records.end(); ++iter)
        {
            if (iter->second.operation_type == DirtyActivationRecord::OperationType::OPERATION_TYPE_DEL)
            {
                database->delete_activation_record(iter->second.activation_record->machine_code);
            }
            else
            {
                database->update_activation_record(iter->second.activation_record);
            }
        }
    }
}

GDate LicensePoints::generate_expiry_time_by_period(int warranty_days, int warranty_months, int warranty_years)
{
    GDate expire_time = Base::TimeUtils::get_current_gdate();

    g_date_add_days(&expire_time, warranty_days);
    g_date_add_months(&expire_time, warranty_months);
    g_date_add_years(&expire_time, warranty_years);

    return expire_time;
}

bool LicensePoints::activations_add_period(int warranty_days, int warranty_months, int warranty_years)
{
    GDate expiry_date = Base::TimeUtils::get_current_gdate();
    ActivationRecordVec activation_records;

    for (auto iter = this->activation_records_.begin(); iter != this->activation_records_.end(); ++iter)
    {
        auto expiry_timestamp = Base::TimeUtils::str2timestamp(iter->second->expiry_date.c_str());
        if ((expiry_timestamp + DAY_IN_SECONDS - 1) > time(NULL))
        {
            expiry_date = Base::TimeUtils::timestamp2gdate(expiry_timestamp);
        }

        g_date_add_days(&expiry_date, warranty_days);
        g_date_add_months(&expiry_date, warranty_months);
        g_date_add_years(&expiry_date, warranty_years);

        iter->second->expiry_date = Base::TimeUtils::gdate2str(&expiry_date);
        auto error_code = this->update_activation_code(iter->second->machine_code, expiry_date, iter->second->activation_code);
        if (error_code != LicenseErrorCode::LEC_SUCCESS)
        {
            KLOG_ERROR(Base::LicenseError::get_error_desc(error_code).c_str());
            return false;
        }
        activation_records.push_back(iter->second);
    }

    this->cache_dirty_records(DirtyActivationRecord::OperationType::OPERATION_TYPE_ADD, activation_records);

    return true;
}

bool LicensePoints::activations_update_expiry_time(GDate expiry_time)
{
    ActivationRecordVec activation_records;

    for (auto iter = this->activation_records_.begin(); iter != this->activation_records_.end(); ++iter)
    {
        iter->second->expiry_date = Base::TimeUtils::gdate2str(&expiry_time);
        auto error_code = this->update_activation_code(iter->second->machine_code, expiry_time, iter->second->activation_code);
        if (error_code != LicenseErrorCode::LEC_SUCCESS)
        {
            KLOG_ERROR(Base::LicenseError::get_error_desc(error_code).c_str());
            return false;
        }
        activation_records.push_back(iter->second);
    }

    this->cache_dirty_records(DirtyActivationRecord::OperationType::OPERATION_TYPE_ADD, activation_records);

    return true;
}

}  // namespace Server
