/**
 * @file          /ks-license/src/server/license-points.h
 * @brief
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved.
 */

#pragma once

#include <set>
#include <thread>
#include "src/server/license-database.h"

namespace Server
{
// 记录需要往数据库同步的信息

struct DirtyActivationRecord
{
    enum OperationType
    {
        OPERATION_TYPE_DEL,
        OPERATION_TYPE_ADD,
        OPERATION_TYPE_MOD,
    };
    OperationType operation_type;
    std::shared_ptr<ActivationRecord> activation_record;
};

using DirtyActivationRecordVec = std::vector<std::shared_ptr<DirtyActivationRecord>>;

struct DirtyPointsInfo
{
    std::shared_ptr<PointsRecord> points_record;
    std::map<std::string, DirtyActivationRecord> activation_records;
};

class LicensePoints
{
public:
    LicensePoints(std::shared_ptr<PointsRecord> points_record);
    virtual ~LicensePoints(){};

    bool init();

    std::string get_uid() { return this->points_record_->uid; }

    // 注意：如果是struct，对外只能获取数据拷贝，否则会出现线程安全问题

    // 获取点数记录
    std::shared_ptr<PointsRecord> copy_points_record();
    bool set_points_record(std::shared_ptr<PointsRecord> points_record);
    // 获取激活信息
    ActivationRecordVec copy_activation_records();
    bool activations_add_period(int warranty_days, int warranty_months, int warranty_years);
    bool activations_update_expiry_time(GDate expiry_time);
    // 通过机器码生成授权码
    LicenseErrorCode generate_activation_code(const std::string &machine_code,
                                              const std::string &client_ip,
                                              std::string &out_license_code);

    // 判断机器码是否已经授权
    bool exist_machine_code(const std::string &machine_code);

    // 更新机器的保活时间
    LicenseErrorCode update_keep_alive(const std::string &machine_code);

private:
    ActivationRecordVec get_activation_records();
    std::shared_ptr<ActivationRecord> get_activation_record(const std::string &machine_code);

    void print_points();

    // 处理激活记录回收
    void process_activation_recycling();
    // 缓存变更记录
    void cache_dirty_records(DirtyActivationRecord::OperationType operation_type, const ActivationRecordVec &activation_records);
    // 同步数据到数据库
    void flush_to_database();

    GDate generate_expiry_time_by_period(int warranty_days, int warranty_months, int warranty_years);

    // 通过机器和过期时间生成激活码:用于更新激活记录，机器码必须要存在
    LicenseErrorCode update_activation_code(const std::string &machine_code_str,
                                            GDate expiry_date,
                                            std::string &out_license_code);

private:
    std::shared_ptr<PointsRecord> points_record_;
    // 激活记录
    std::map<std::string, std::shared_ptr<ActivationRecord>> activation_records_;

    // 同步数据库的缓存记录
    bool is_dirty_;
    DirtyPointsInfo dirty_points_info_;
    std::recursive_mutex mutex_;
    std::mutex dirty_mutex_;
    std::mutex flush_db_mutex_;
};

typedef std::vector<std::shared_ptr<LicensePoints>> LicensePointsVec;

}  // namespace Server